import os
import signal
import sys
import time
import socket
import subprocess
import psutil
from contextlib import closing
from typing import Optional

class ServerManager:
    def __init__(self, app_path: str, host: str, port: int, name: str):
        self.app_path = app_path
        self.host = host
        self.port = port
        self.name = name
        self._process: Optional[subprocess.Popen] = None
        self._pid_file = os.path.join(os.path.expanduser("~"), f".upsonic_{name}_server.pid")

    def _kill_process_using_port(self) -> bool:
        """Find and kill all processes using the specified port."""
        killed_any = False
        failed_kills = []
        
        for proc in psutil.process_iter(['pid', 'name']):
            try:
                # Get connections separately since it can't be retrieved in process_iter
                process = psutil.Process(proc.info['pid'])
                connections = process.net_connections()
                for conn in connections:
                    if hasattr(conn, 'laddr') and len(conn.laddr) >= 2 and conn.laddr[1] == self.port:
                        # Found a process using our port
                        try:
                            process_info = f"PID: {process.pid}, Name: {process.name()}"
                            # Try SIGTERM first
                            process.terminate()
                            try:
                                process.wait(timeout=3)
                                killed_any = True
                            except psutil.TimeoutExpired:
                                # If SIGTERM doesn't work, use SIGKILL
                                try:
                                    process.kill()
                                    process.wait(timeout=1)
                                    killed_any = True
                                except Exception as e:
                                    failed_kills.append(f"{process_info} (Kill failed: {str(e)})")
                        except (psutil.NoSuchProcess, psutil.AccessDenied) as e:
                            try:
                                # Last resort: try using system kill command
                                os.kill(process.pid, signal.SIGKILL)
                                killed_any = True
                            except Exception as e:
                                failed_kills.append(f"{process_info} (System kill failed: {str(e)})")
            except (psutil.NoSuchProcess, psutil.AccessDenied, psutil.TimeoutExpired) as e:
                continue
            except Exception as e:
                continue

        if failed_kills:
            error_msg = f"Failed to kill the following processes using port {self.port}:\n" + "\n".join(failed_kills)
            raise RuntimeError(error_msg)
            
        return killed_any

    def _is_port_in_use(self, kill_if_used: bool = False) -> bool:
        """Check if the port is in use and optionally kill the process using it."""
        with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as sock:
            result = sock.connect_ex((self.host, self.port)) == 0
            if result and kill_if_used:
                return not self._kill_process_using_port()
            return result

    def _write_pid(self):
        """Write the current process ID to the PID file"""
        if self._process and self._process.pid:
            with open(self._pid_file, 'w') as f:
                f.write(str(self._process.pid))

    def _read_pid(self) -> Optional[int]:
        """Read the process ID from the PID file"""
        try:
            if os.path.exists(self._pid_file):
                with open(self._pid_file, 'r') as f:
                    return int(f.read().strip())
        except (ValueError, IOError):
            pass
        return None

    def _cleanup_pid(self):
        """Remove the PID file"""
        try:
            if os.path.exists(self._pid_file):
                os.remove(self._pid_file)
        except OSError:
            pass

    def _force_cleanup_port(self):
        """Aggressively clean up the port before starting."""
        if not self._is_port_in_use():
            return

        # First try normal port kill
        try:
            self._kill_process_using_port()
            # Wait a bit to ensure port is freed
            time.sleep(1)
        except RuntimeError as e:
            # If normal kill failed, try alternative method using psutil
            try:
                killed_processes = []
                for proc in psutil.process_iter(['pid', 'name']):
                    try:
                        process = psutil.Process(proc.info['pid'])
                        connections = process.net_connections(kind='inet')
                        for conn in connections:
                            # Check if the connection is using our port
                            if len(conn.laddr) >= 2 and conn.laddr[1] == self.port:
                                # Try SIGTERM first
                                process.terminate()
                                try:
                                    process.wait(timeout=3)
                                    killed_processes.append(proc.info['pid'])
                                except psutil.TimeoutExpired:
                                    # If SIGTERM doesn't work, use SIGKILL
                                    process.kill()
                                    process.wait(timeout=1)
                                    killed_processes.append(proc.info['pid'])
                    except (psutil.NoSuchProcess, psutil.AccessDenied, psutil.TimeoutExpired):
                        continue
                    except Exception:
                        continue

                if not killed_processes:
                    # If no processes were found using psutil's connections method,
                    # try one last attempt using socket connections
                    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                    try:
                        # Try to bind to the port
                        sock.bind((self.host, self.port))
                    except socket.error:
                        # If we can't bind, try to find what's using it
                        for proc in psutil.process_iter(['pid', 'name']):
                            try:
                                process = psutil.Process(proc.info['pid'])
                                if any(len(conn.laddr) >= 2 and conn.laddr[1] == self.port 
                                      for conn in process.net_connections(kind='inet')):
                                    process.kill()
                                    killed_processes.append(proc.info['pid'])
                            except (psutil.NoSuchProcess, psutil.AccessDenied):
                                continue
                    finally:
                        sock.close()

                time.sleep(1)
                
            except Exception as e:
                raise RuntimeError(f"Unexpected error while cleaning up port {self.port}: {str(e)}")

        # Final check
        if self._is_port_in_use():
            raise RuntimeError(f"Port {self.port} is still in use after all cleanup attempts. Please check running processes manually.")

    def start(self, redirect_output: bool = False, force: bool = False):
        """Start the server if it's not already running."""
        # Always do an initial cleanup
        self._force_cleanup_port()
        
        if self.is_running():
            return

        # Check if port is still in use after cleanup
        if self._is_port_in_use(kill_if_used=force):
            raise RuntimeError(f"Port {self.port} is still in use and could not be freed")

        # Set up logging
        if redirect_output:
            log_dir = "logs"
            if not os.path.exists(log_dir):
                os.makedirs(log_dir)
            stdout = open(os.path.join(log_dir, f'{self.name}_server.log'), 'a')
            stderr = open(os.path.join(log_dir, f'{self.name}_server_error.log'), 'a')
        else:
            stdout = None
            stderr = None

        # Prepare the command
        cmd = [
            sys.executable, "-m", "uvicorn",
            self.app_path,
            "--host", self.host,
            "--port", str(self.port),
            "--log-level", "error",
            "--no-access-log"
        ]

        try:
            # Start the server process
            self._process = subprocess.Popen(
                cmd,
                stdout=stdout,
                stderr=stderr,
                start_new_session=True  # This creates a new process group
            )
            self._write_pid()

            # Wait for server to start
            start_time = time.time()
            max_wait_time = 30  # Increase wait time for slower systems
            while not self._is_port_in_use() and time.time() - start_time < max_wait_time:
                if self._process.poll() is not None:
                    # Process has terminated
                    raise RuntimeError(f"Server process terminated unexpectedly with code {self._process.returncode}")
                time.sleep(0.1)

            if time.time() - start_time >= max_wait_time:
                raise RuntimeError(f"Timeout waiting for {self.name} server to start")

        except Exception as e:
            self.stop()  # Clean up if startup fails
            raise RuntimeError(f"Failed to start {self.name} server: {str(e)}")

    def stop(self):
        """Stop the server if it's running."""
        pid = self._read_pid()
        if pid:
            try:
                process = psutil.Process(pid)
                # Try to terminate the process group
                os.killpg(os.getpgid(process.pid), signal.SIGTERM)
                try:
                    process.wait(timeout=5)
                except psutil.TimeoutExpired:
                    # If termination fails, force kill the process group
                    os.killpg(os.getpgid(process.pid), signal.SIGKILL)
            except (psutil.NoSuchProcess, ProcessLookupError):
                pass

        if self._process:
            try:
                self._process.terminate()
                self._process.wait(timeout=5)
            except (subprocess.TimeoutExpired, ProcessLookupError):
                if self._process.poll() is None:
                    self._process.kill()

        self._process = None
        self._cleanup_pid()

    def is_running(self) -> bool:
        """Check if the server is currently running."""
        if self._process and self._process.poll() is None:
            return True

        pid = self._read_pid()
        if pid:
            try:
                process = psutil.Process(pid)
                return process.is_running() and process.name().startswith("python")
            except psutil.NoSuchProcess:
                self._cleanup_pid()
        return False 


from ..server_manager import ServerManager
from multiprocessing import freeze_support

_server_manager = ServerManager(
    app_path="upsonic.tools_server.server.api:app",
    host="localhost",
    port=8086,
    name="tools"
)

def run_tools_server(redirect_output: bool = False):
    """Start the tools server if it's not already running."""
    _server_manager.start(redirect_output=redirect_output)

def run_tools_server_internal(reload: bool = True):
    """Run the tools server directly (for development)"""
    import uvicorn
    uvicorn.run("upsonic.tools_server.server.api:app", host="localhost", port=8086, reload=reload)

def stop_tools_server():
    """Stop the tools server if it's running."""
    _server_manager.stop()

def is_tools_server_running() -> bool:
    """Check if the tools server is currently running."""
    return _server_manager.is_running()

if __name__ == '__main__':
    freeze_support()

__all__ = ["run_tools_server", "stop_tools_server", "is_tools_server_running", "app", "run_tools_server_internal"]



from copy import deepcopy
from typing import Any, Optional, Union, Type
from .client.tasks.tasks import Task

from .client.agent_configuration.agent_configuration import AgentConfiguration

from pydantic import Field

from .client.tasks.task_response import ObjectResponse


class ValidationResult(ObjectResponse):
    any_suspicion: bool
    feedback: str

validator_task_prompt = """
Evaluate the previous question and its answer for accuracy, consistency, and completeness. Also consider the requested output format. Be extremely vigilant for any non-referenced values or unsupported claims. Pay particular attention to:

Unverified information: Identify and flag any claims, statistics, or statements that lack credible sources or proper citations.

Unsupported assertions: Watch for any information presented as fact without adequate evidence or references.

Vague attributions: Be wary of phrases like "Studies show..." or "Experts say..." without specific, verifiable sources.

Misuse of data: Check for any misinterpretation or misapplication of statistical information.

Temporal inconsistencies: Ensure all historical claims or dates are properly verified and sourced.

Speculative content: Identify any conjectures or hypothetical scenarios presented without clear indication of their speculative nature.

Examples of issues to flag:

Any statistic or numerical claim without a specific, credible source
Historical events or dates mentioned without verifiable references
Scientific claims or breakthroughs without links to peer-reviewed research
Expert opinions quoted without naming the expert and their credentials
Trends or patterns described without supporting data
Do not allow any unverified information, guesses, or assumptions to pass unchallenged. Rigorously check all data and its sources. If there is any information without a valuable and meaningful reference, highlight it for removal or verification. Ensure all information is factual, current, and properly supported by credible, cited sources.

"""


editor_task_prompt = """
As the editor agent, your task is to review and refine the output provided. Pay close attention to the issues highlighted in the previous evaluation. Your primary responsibilities are:

Remove or correct any misunderstandings, unverified information, or hallucinations identified in the output.

Ensure all claims are supported by credible, up-to-date sources.

Eliminate any statements that present personal opinions as facts.

Verify all statistical data, historical dates, and scientific claims.

Check for and remove any mentions of non-existent technologies, people, or places.

Maintain the requested output format while improving accuracy and reliability.

Remember, you should not return any data that doesn't meet these high standards of verification and accuracy. If you cannot verify a piece of information or find a reliable source for it, remove it from the output.

Your goal is to produce a refined version of the output that is factually accurate, consistently reliable, and complete within the bounds of verified information. Quality and accuracy are paramount - it's better to have less information that is fully verified than more information that includes questionable data.

"""

class ReliabilityProcessor:
    @staticmethod
    def process_result(
        result: Any, 
        reliability_layer: Optional[Any] = None, 
        task: Optional[Task] = None,
        llm_model: Optional[str] = None
    ) -> Any:
        """
        Process the result based on reliability layer settings.
        If reliability_layer is None or fields are 0, return original result.
        
        Args:
            result: The result to process
            reliability_layer: Configuration for reliability checks (can be instance or class)
            task: The original task that generated this result
            llm_model: The LLM model used to generate the result
        """
        if reliability_layer is None:
            return result

        # Get prevent_hallucination value
        prevent_hallucination = getattr(reliability_layer, 'prevent_hallucination', 0)
        if isinstance(prevent_hallucination, property):
            prevent_hallucination = prevent_hallucination.fget(reliability_layer)

        processed_result = result

        # Check prevent_hallucination
        if prevent_hallucination > 0:

            if prevent_hallucination == 10:
                validator_agent = AgentConfiguration("Information Validator Agent", model=llm_model, sub_task=False)
                copy_task = deepcopy(task)
                copy_task._response = result
                validator_task = Task(validator_task_prompt, context=[copy_task, copy_task.response_format], response_format=ValidationResult, tools=task.tools)
                validator_agent.do(validator_task)

                if validator_task.response.any_suspicion == False:
                    return result
                else:
                    editor_agent = AgentConfiguration("Information Editor Agent", model=llm_model, sub_task=False)
                    editor_task = Task(editor_task_prompt, context=[copy_task, copy_task.response_format, validator_task, str(validator_task.response.model_dump(mode="json"))], response_format=task.response_format, tools=task.tools)
                    editor_agent.do(editor_task)

                    return editor_task.response



        return processed_result 


from typing import Any
from rich.console import Console
from rich.panel import Panel
from rich.table import Table
from rich.align import Align
from .price import get_estimated_cost



console = Console()

def spacing():
    console.print("")



def connected_to_server(server_type: str, status: str):
    """
    Prints a 'Connected to Server' section for Upsonic, full width,
    with two columns: 
      - left column (labels) left-aligned
      - right column (values) left-aligned, positioned on the right half 
    """

    # Determine color and symbol for the status
    if status.lower() == "established":
        status_text = "[green]✓ Established[/green]"
    elif status.lower() == "failed":
        status_text = "[red]✗ Failed[/red]"
    else:
        status_text = f"[cyan]… {status}[/cyan]"

    # Build a table that expands to full console width
    table = Table(show_header=False, expand=True, box=None)
    
    # We define 2 columns, each with ratio=1 so they evenly split the width
    # Both are left-aligned, but the second column will end up on the right half.
    table.add_column("Label", justify="left", ratio=1)
    table.add_column("Value", justify="left", ratio=1)

    # Rows: one for server type, one for status
    table.add_row("[bold]Server Type:[/bold]", f"[yellow]{server_type}[/yellow]")
    table.add_row("[bold]Connection Status:[/bold]", status_text)

    table.width = 60

    # Wrap the table in a Panel that also expands full width
    panel = Panel(
        table, 
        title="[bold cyan]Upsonic - Server Connection[/bold cyan]",
        border_style="cyan",
        expand=True,  # panel takes the full terminal width
        width=70  # Adjust as preferred
    )

    # Print the panel (it will fill the entire width, with two columns inside)
    console.print(panel)

    spacing()

def call_end(result: Any, llm_model: str, response_format: str, start_time: float, end_time: float, usage: dict, debug: bool = False):
    table = Table(show_header=False, expand=True, box=None)
    table.width = 60

    table.add_row("[bold]LLM Model:[/bold]", f"{llm_model}")
    # Add spacing
    table.add_row("")

    from ..client.level_two.agent import SubTaskList, SearchResult, CompanyObjective, HumanObjective

    is_it_subtask = isinstance(result, SubTaskList)
    is_it_search = isinstance(result, SearchResult)
    is_it_company = isinstance(result, CompanyObjective)
    is_it_human = isinstance(result, HumanObjective)

    if is_it_subtask:
        # Print total task count
        table.add_row(f"[bold]Total Subtasks:[/bold]", f"[yellow]{len(result.sub_tasks)}[/yellow]")
        table.add_row("")
        # Print each task as well as bullet list
        for each in result.sub_tasks:
            table.add_row(f"[bold]Subtask:[/bold]", f"[green]{each.description}[/green]")
            table.add_row(f"[bold]Required Output:[/bold]", f"[green]{each.required_output}[/green]")
            table.add_row(f"[bold]Tools:[/bold]", f"[green]{each.tools}[/green]")
            table.add_row("")
    elif is_it_search:
        table.add_row("[bold]Has Customers:[/bold]", f"[green]{'Yes' if result.any_customers else 'No'}[/green]")
        table.add_row("")
        table.add_row("[bold]Products:[/bold]")
        for product in result.products:
            table.add_row("", f"[green]• {product}[/green]")
        table.add_row("")
        table.add_row("[bold]Services:[/bold]")
        for service in result.services:
            table.add_row("", f"[green]• {service}[/green]")
        table.add_row("")
        table.add_row("[bold]Potential Competitors:[/bold]")
        for competitor in result.potential_competitors:
            table.add_row("", f"[yellow]• {competitor}[/yellow]")
        table.add_row("")
    elif is_it_company:
        table.add_row("[bold]Company Objective:[/bold]", f"[blue]{result.objective}[/blue]")
        table.add_row("")
        table.add_row("[bold]Goals:[/bold]")
        for goal in result.goals:
            table.add_row("", f"[blue]• {goal}[/blue]")
        table.add_row("")
        table.add_row("[bold]State:[/bold]", f"[blue]{result.state}[/blue]")
        table.add_row("")
    elif is_it_human:
        table.add_row("[bold]Job Title:[/bold]", f"[magenta]{result.job_title}[/magenta]")
        table.add_row("")
        table.add_row("[bold]Job Description:[/bold]", f"[magenta]{result.job_description}[/magenta]")
        table.add_row("")
        table.add_row("[bold]Job Goals:[/bold]")
        for goal in result.job_goals:
            table.add_row("", f"[magenta]• {goal}[/magenta]")
        table.add_row("")
    else:
        result_str = str(result)
        # Limit result to 370 characters
        if not debug:
            result_str = result_str[:370]
        # Add ellipsis if result is truncated
        if len(result_str) < len(str(result)):
            result_str += "[bold white]...[/bold white]"

        table.add_row("[bold]Result:[/bold]", f"[green]{result_str}[/green]")

    # Add spacing
    table.add_row("")
    table.add_row("[bold]Response Format:[/bold]", f"{response_format}")

    table.add_row("[bold]Estimated Cost:[/bold]", f"{get_estimated_cost(usage['input_tokens'], usage['output_tokens'], llm_model)}$")
    time_taken = end_time - start_time
    time_taken_str = f"{time_taken:.2f} seconds"
    table.add_row("[bold]Time Taken:[/bold]", f"{time_taken_str}")
    panel = Panel(
        table,
        title="[bold white]Upsonic - Call Result[/bold white]",
        border_style="white",
        expand=True,
        width=70
    )

    console.print(panel)
    spacing()



def agent_end(result: Any, llm_model: str, response_format: str, start_time: float, end_time: float, usage: dict, tool_count: int, context_count: int, debug: bool = False):
    table = Table(show_header=False, expand=True, box=None)
    table.width = 60

    table.add_row("[bold]LLM Model:[/bold]", f"{llm_model}")
    # Add spacing
    table.add_row("")
    result_str = str(result)
    # Limit result to 370 characters
    if not debug:
        result_str = result_str[:370]
    # Add ellipsis if result is truncated
    if len(result_str) < len(str(result)):
        result_str += "[bold white]...[/bold white]"

    table.add_row("[bold]Result:[/bold]", f"[green]{result_str}[/green]")
    # Add spacing
    table.add_row("")
    table.add_row("[bold]Response Format:[/bold]", f"{response_format}")
    
    table.add_row("[bold]Tools:[/bold]", f"{tool_count} [bold]Context Used:[/bold]", f"{context_count}")
    table.add_row("[bold]Estimated Cost:[/bold]", f"{get_estimated_cost(usage['input_tokens'], usage['output_tokens'], llm_model)}$")
    time_taken = end_time - start_time
    time_taken_str = f"{time_taken:.2f} seconds"
    table.add_row("[bold]Time Taken:[/bold]", f"{time_taken_str}")
    panel = Panel(
        table,
        title="[bold white]Upsonic - Agent Result[/bold white]",
        border_style="white",
        expand=True,
        width=70
    )

    console.print(panel)
    spacing()


def agent_total_cost(total_input_tokens: int, total_output_tokens: int, total_time: float, llm_model: str):
    table = Table(show_header=False, expand=True, box=None)
    table.width = 60

    table.add_row("[bold]Estimated Cost:[/bold]", f"{get_estimated_cost(total_input_tokens, total_output_tokens, llm_model)}$")
    table.add_row("[bold]Time Taken:[/bold]", f"{total_time:.2f} seconds")
    panel = Panel(
        table,
        title="[bold white]Upsonic - Agent Total Cost[/bold white]",
        border_style="white",
        expand=True,
        width=70
    )
    console.print(panel)
    spacing()

def agent_retry(retry_count: int, max_retries: int):
    table = Table(show_header=False, expand=True, box=None)
    table.width = 60

    table.add_row("[bold]Retry Status:[/bold]", f"[yellow]Attempt {retry_count + 1} of {max_retries + 1}[/yellow]")
    
    panel = Panel(
        table,
        title="[bold yellow]Upsonic - Agent Retry[/bold yellow]",
        border_style="yellow",
        expand=True,
        width=70
    )

    console.print(panel)
    spacing()


from fastapi import HTTPException
from pydantic import BaseModel
from typing import List, Dict, Any, Optional, Union
import traceback
from ...api import app, timeout
from ..call import Call
import asyncio
import cloudpickle
cloudpickle.DEFAULT_PROTOCOL = 2
import base64


prefix = "/level_one"


class GPT4ORequest(BaseModel):
    prompt: str
    response_format: Optional[Any] = []
    tools: Optional[Any] = []
    context: Optional[Any] = None
    llm_model: Optional[Any] = "openai/gpt-4o"
    system_prompt: Optional[Any] = None


@app.post(f"{prefix}/gpt4o")
@timeout(300.0)  # 5 minutes timeout for AI operations
async def call_gpt4o(request: GPT4ORequest):
    """
    Endpoint to call GPT-4 with optional tools and MCP servers.

    Args:
        request: GPT4ORequest containing prompt and optional parameters

    Returns:
        The response from the AI model
    """
    try:
        # Handle pickled response format
        if request.response_format != "str":
            try:
                # Decode and unpickle the response format
                pickled_data = base64.b64decode(request.response_format)
                response_format = cloudpickle.loads(pickled_data)
            except Exception as e:
                traceback.print_exc()
                # Fallback to basic type mapping if unpickling fails
                type_mapping = {
                    "str": str,
                    "int": int,
                    "float": float,
                    "bool": bool,
                }
                response_format = type_mapping.get(request.response_format, str)
        else:
            response_format = str

        if request.context is not None:
            try:
                pickled_context = base64.b64decode(request.context)
                context = cloudpickle.loads(pickled_context)
            except Exception as e:
                traceback.print_exc()
                context = None
        else:
            context = None

        result = await Call.gpt_4o(
            prompt=request.prompt,
            response_format=response_format,
            tools=request.tools,
            context=context,
            llm_model=request.llm_model,
            system_prompt=request.system_prompt
        )

        if request.response_format != "str" and result["status_code"] == 200:
            result["result"] = cloudpickle.dumps(result["result"])
            result["result"] = base64.b64encode(result["result"]).decode('utf-8')
        return {"result": result, "status_code": 200}
    except Exception as e:
        traceback.print_exc()
        return {"result": {"status_code": 500, "detail": f"Error processing Call request: {str(e)}"}, "status_code": 500}



from dotenv import load_dotenv
load_dotenv()

from ....storage.configuration import Configuration

import asyncio
import atexit



class BrowserManager:
    _instance = None
    _browser = None
    _loop = None

    @classmethod
    def get_instance(cls):
        if cls._instance is None:
            cls._instance = cls()
            # Register the cleanup function
            atexit.register(cls._cleanup)
        return cls._instance

    @classmethod
    async def initialize(cls):
        instance = cls.get_instance()
        if instance._browser is None:
            from browser_use import Browser
            browser = Browser()
            instance._browser = browser
            instance._loop = asyncio.get_event_loop()
            return browser
        return instance._browser

    @classmethod
    async def get_context(cls):
        """Get a new browser context for isolation"""
        instance = cls.get_instance()
        if instance._browser:
            context = await instance._browser.new_context()
            return context
        return None

    @classmethod
    def _cleanup(cls):
        """Cleanup function that will be called when the Python process exits"""
        instance = cls.get_instance()
        if instance._browser and instance._loop:
            # Create a new event loop if the main one is closed
            try:
                loop = instance._loop if instance._loop.is_running() else asyncio.new_event_loop()
                asyncio.set_event_loop(loop)
                loop.run_until_complete(instance._browser.close())
            except:
                pass  # Suppress any errors during shutdown

    @classmethod
    async def close(cls):
        """Manual close method - only use if you explicitly need to close the browser"""
        instance = cls.get_instance()
        if instance._browser:
            await instance._browser.close()
            instance._browser = None

    @classmethod
    def get_browser(cls):
        instance = cls.get_instance()
        return instance._browser


class LLMManager:
    _instance = None
    _llm_model = None

    @classmethod
    def get_instance(cls):
        if cls._instance is None:
            cls._instance = cls()
        return cls._instance

    @classmethod
    def set_model(cls, model):
        instance = cls.get_instance()
        instance._llm_model = model
        print("SETTING THE LLM MODEL TO:", model)

    @classmethod
    def get_model(cls):
        instance = cls.get_instance()
        print("GETTING THE LLM MODEL:", instance._llm_model)
        return instance._llm_model


def get_llm():
    llm_model = LLMManager.get_model()
    print("THE LLM MODEL IS", llm_model)
    
    if not llm_model:
        raise ValueError("LLM model not set before calling get_llm()")
    
    # Map our model names to standard model names
    openai_model_mapping = {
        "openai/gpt-4o": "gpt-4o",
        "gpt-4o": "gpt-4o",
        "openai/o3-mini": "o3-mini",
        "openai/gpt-4o-mini": "gpt-4o",
        "azure/gpt-4o": "gpt-4o",
        "gpt-4o-azure": "gpt-4o"
    }

    claude_model_mapping = {
        "claude/claude-3-5-sonnet": "claude-3-5-sonnet-latest",
        "claude-3-5-sonnet": "claude-3-5-sonnet-latest",
        "bedrock/claude-3-5-sonnet": "us.anthropic.claude-3-5-sonnet-20241022-v2:0",
        "claude-3-5-sonnet-aws": "us.anthropic.claude-3-5-sonnet-20241022-v2:0"
    }

    deepseek_model_mapping = {
        "deepseek/deepseek-chat": "deepseek-chat"
    }
    
    # Handle Azure OpenAI
    if llm_model in ["azure/gpt-4o", "gpt-4o-azure"]:
        azure_endpoint = Configuration.get("AZURE_OPENAI_ENDPOINT")
        azure_api_key = Configuration.get("AZURE_OPENAI_API_KEY")
        azure_api_version = Configuration.get("AZURE_OPENAI_API_VERSION", "2024-10-21")
        

            
        from langchain_openai import AzureChatOpenAI
        llm = AzureChatOpenAI(
            model="gpt-4o",
            api_version=azure_api_version,
            azure_endpoint=azure_endpoint,
            api_key=azure_api_key
        )
    
    # Handle regular OpenAI
    elif llm_model in openai_model_mapping:
        openai_api_key = Configuration.get("OPENAI_API_KEY")
        if not openai_api_key:
            raise ValueError("OpenAI API key not found in configuration")
        from langchain_openai import ChatOpenAI
        llm = ChatOpenAI(
            model_name=openai_model_mapping[llm_model],
            openai_api_key=openai_api_key,
        )

    # Handle Claude (Anthropic)
    elif llm_model in claude_model_mapping:
        if llm_model in ["bedrock/claude-3-5-sonnet", "claude-3-5-sonnet-aws"]:
            # AWS Bedrock configuration
            aws_access_key_id = Configuration.get("AWS_ACCESS_KEY_ID")
            aws_secret_access_key = Configuration.get("AWS_SECRET_ACCESS_KEY")
            aws_region = Configuration.get("AWS_REGION")
            
            if not all([aws_access_key_id, aws_secret_access_key, aws_region]):
                raise ValueError("AWS credentials not found in configuration")
            from langchain_community.chat_models import BedrockChat
            llm = BedrockChat(
                model_id=claude_model_mapping[llm_model],
                credentials_profile_name=None,
                region_name=aws_region,
                aws_access_key_id=aws_access_key_id,
                aws_secret_access_key=aws_secret_access_key
            )
        else:
            # Regular Anthropic configuration
            anthropic_api_key = Configuration.get("ANTHROPIC_API_KEY")
            if not anthropic_api_key:
                raise ValueError("Anthropic API key not found in configuration")
            from langchain_anthropic import ChatAnthropic
            llm = ChatAnthropic(
                model=claude_model_mapping[llm_model],
                anthropic_api_key=anthropic_api_key,
                temperature=0.0,
                timeout=100
            )

    # Handle DeepSeek models
    elif llm_model in deepseek_model_mapping:
        deepseek_api_key = Configuration.get("DEEPSEEK_API_KEY")
        if not deepseek_api_key:
            raise ValueError("DeepSeek API key not found in configuration")
            
        from langchain_openai import ChatOpenAI
        llm = ChatOpenAI(
            model_name=deepseek_model_mapping[llm_model],
            api_key=deepseek_api_key,
            base_url="https://api.deepseek.com/v1"
        )
    
    else:
        raise ValueError(f"Unsupported model for browser use: {llm_model}")
    
    return llm


async def BrowserUse__browser_agent(task: str, expected_output: str):
    """An AI agent that can browse the web, extract information, and perform actions."""
    from browser_use import Agent
    
    # Get or create the browser instance
    browser = await BrowserManager.initialize()
    
    # Create a new context for this agent run
    context = await BrowserManager.get_context()
    
    # Create the agent with the browser context
    agent = Agent(
        task=task+"\n\nExpected Output: "+expected_output,
        llm=get_llm(),
        browser=browser,
        browser_context=context,  # Use persistent context
        generate_gif=False
    )
    
    try:
        result = await agent.run()
        return result.final_result()
    finally:
        # Clean up the context after the agent is done
        if context:
            await context.close()


# List of all browser use tools
BrowserUse_tools = [
    BrowserUse__browser_agent
]



import inspect
import cloudpickle

from ..level_utilized.utility import error_handler
cloudpickle.DEFAULT_PROTOCOL = 2
import dill
import base64
import httpx
from typing import Any, List, Dict, Optional, Type, Union, Callable
from pydantic import BaseModel
from functools import wraps

import inspect
import functools

from ..tasks.tasks import Task
from ...exception import NoAPIKeyException, UnsupportedLLMModelException

class ComputerUse:
    pass

class BrowserUse:
    pass

class Search:
    pass


def generate_static_method_class(instance):


    
    # Store instance attributes
    instance_attrs = {name: value for name, value in inspect.getmembers(instance)
                     if not name.startswith('__') and not callable(value)}
    
    # Create new class with the same name as the original class
    original_class_name = type(instance).__name__
    TransformedClass = type(original_class_name, (), {})
    
    # Set instance attributes as class attributes
    for attr_name, attr_value in instance_attrs.items():
        setattr(TransformedClass, attr_name, attr_value)
    


    # Dynamically add each method as static method to the new class
    for method_name, method in inspect.getmembers(instance, predicate=inspect.ismethod):

        if not method_name.startswith('__'):

            # Create a closure that captures the instance attributes
            def create_static_method(method, instance_attrs):
                @functools.wraps(method)
                def static_wrapper(*args, **kwargs):
                    # Create a new instance with the stored attributes
                    temp_instance = type(instance)(**{k: v for k, v in instance_attrs.items()})
                    return method.__get__(temp_instance, type(instance))(*args, **kwargs)
                return static_wrapper

            static_method = staticmethod(create_static_method(method, instance_attrs))
            setattr(TransformedClass, method_name, static_method)


    return TransformedClass

class Tools:
    def tool(self, library: Optional[Union[str, List[str]]] = None):
        """
        Decorator to register a function or class as a tool.
        Can be used as @tool(), @tool("pandas"), or @tool(["pandas", "numpy"])

        Args:
            library: Optional library name or list of library names to install before registering the tool
        """
        def decorator(obj: Union[Callable, Type]):
            # Install libraries first if specified
            if library:
                if isinstance(library, str):
                    self.install_library(library)
                else:
                    for lib in library:
                        self.install_library(lib)
            
            # If it's a class, register each method as a tool
            if isinstance(obj, type):
                class_name = obj.__name__
                # Get all methods that don't start with underscore
                methods = [(name, getattr(obj, name)) for name in dir(obj) 
                          if not name.startswith('_') and callable(getattr(obj, name))]
                
                # Register each method as a tool
                for name, method in methods:
                    # Convert the method to a standalone function
                    def create_standalone(method, full_name):
                        @wraps(method)
                        def standalone(*args, **kwargs):
                            return method(*args, **kwargs)
                        standalone.__name__ = full_name
                        return standalone
                    
                    full_name = f"{class_name}__{name}"
                    standalone = create_standalone(method, full_name)
                    self.add_tool(standalone)
                
                return obj
            
            if isinstance(obj, object):
                obj = generate_static_method_class(obj)


                # Get all methods that don't start with underscore
                methods = [(name, getattr(obj, name)) for name in dir(obj) 
                          if not name.startswith('_') and callable(getattr(obj, name))]
                
                # Register each method as a tool
                for name, method in methods:
                    # Convert the method to a standalone function
                    def create_standalone(method, full_name):
                        @wraps(method)
                        def standalone(*args, **kwargs):
                            return method(*args, **kwargs)
                        standalone.__name__ = full_name
                        return standalone
                    
                    full_name = f"{obj.__name__}__{name}"
                    standalone = create_standalone(method, full_name)
                    self.add_tool(standalone)   
                
            else:
                # Register the function as a tool
                @wraps(obj)
                def wrapper(*args, **kwargs):
                    return obj(*args, **kwargs)
                self.add_tool(wrapper)
                return wrapper
                
        return decorator

    def add_tool(
        self,
        function,
    ) -> Any:



        # Get the function then make a cloudpickle of it
        the_module = dill.detect.getmodule(function)
        if the_module is not None:
            cloudpickle.register_pickle_by_value(the_module)

        the_dumped_function = cloudpickle.dumps(function)

        data = {
            "function": base64.b64encode(the_dumped_function).decode("utf-8"),
        }
        
        result = self.send_request("/tools/add_tool", data)
        return result
    


    def add_mcp_tool(self, name: str, command: str, args: List[str], env: Dict[str, str] = {}) -> Dict[str, Any]:
        print("********* ADDING MCP TOOL *********")
        result = self.send_request("/tools/add_mcp_tool", {"name": name, "command": command, "args": args, "env": env})
        error_handler(result)
        print(result)
        print("********* MCP TOOL ADDED *********")
        return result

    def install_library(self, library: str) -> Dict[str, Any]:
        result = self.send_request("/tools/install_library", {"library": library})
        return result

    def uninstall_library(self, library: str) -> Dict[str, Any]:
        result = self.send_request("/tools/uninstall_library", {"library": library})
        return result

    def mcp(self):
        """
        Decorator to register a class as an MCP tool.
        Usage:
        @client.mcp()
        class ToolName:
            command = "command-name"
            args = ["arg1", "arg2"]
            env = {"key": "value"}
        """
        def decorator(cls):
            command = getattr(cls, "command", None)
            args = getattr(cls, "args", [])
            env = getattr(cls, "env", {})

            name = cls.__name__
            
            if not command:
                raise ValueError("MCP tool class must have a 'command' attribute")
                
            self.add_mcp_tool(name, command, args, env)
            return cls
        return decorator



import warnings

warnings.filterwarnings("ignore", category=ResourceWarning)
warnings.filterwarnings("ignore", category=PendingDeprecationWarning)
warnings.filterwarnings("ignore", category=DeprecationWarning)


from .client.base import UpsonicClient
from .client.tasks.task_response import ObjectResponse, StrResponse, IntResponse, FloatResponse, BoolResponse, StrInListResponse
from .client.tasks.tasks import Task
from .client.agent_configuration.agent_configuration import AgentConfiguration
from .client.agent_configuration.agent_configuration import AgentConfiguration as Agent
from .client.knowledge_base.knowledge_base import KnowledgeBase
from .client.direct_llm_call.direct_llm_cal import Direct
from .client.multi_agent.multi_agent import MultiAgent


from .client.storage.storage import ClientConfig

from pydantic import Field


def hello() -> str:
    return "Hello from upsonic!"


__all__ = ["hello", "UpsonicClient", "ObjectResponse", "StrResponse", "IntResponse", "FloatResponse", "BoolResponse", "Task", "StrInListResponse", "AgentConfiguration", "Field", "KnowledgeBase", "ClientConfig", "Agent", "Direct", "MultiAgent"]



import pytest
import pytest_asyncio
from fastapi.testclient import TestClient
from upsonic.tools_server.server.api import (
    app,
    timeout,
    timeout_handler,
    TimeoutException,
)
import asyncio
from httpx import AsyncClient

# Create a new event loop for each test
@pytest.fixture
def event_loop():
    loop = asyncio.new_event_loop()
    yield loop
    loop.close()

# Create an async client for testing
@pytest_asyncio.fixture
async def async_client():
    async with AsyncClient(app=app, base_url="http://test") as client:
        yield client

@pytest.mark.asyncio
async def test_timeout_handler():
    # Test successful completion within timeout
    async def quick_operation():
        return "success"

    result = await timeout_handler(1.0, quick_operation())
    assert result == "success"

    # Test timeout exception
    async def slow_operation():
        await asyncio.sleep(2)
        return "too late"

    with pytest.raises(TimeoutException):
        await timeout_handler(1.0, slow_operation())

@pytest.mark.asyncio
async def test_timeout_decorator():
    async def test_slow_endpoint():
        await asyncio.sleep(2)
        return {"message": "success"}

    # Test that the timeout works using asyncio.timeout
    with pytest.raises(asyncio.TimeoutError):
        async with asyncio.timeout(1.0):
            await test_slow_endpoint()

    # Test successful completion
    async def test_quick_endpoint():
        await asyncio.sleep(0.1)
        return {"message": "success"}

    async with asyncio.timeout(2.0):
        result = await test_quick_endpoint()
        assert result == {"message": "success"}

@pytest.mark.asyncio
async def test_status_endpoint(async_client):
    response = await async_client.get("/status")
    assert response.status_code == 200
    assert response.json() == {"status": "Server is running"}









import traceback
import anthropic
import openai
from pydantic import BaseModel


from typing import Any, Optional

from ...storage.configuration import Configuration

from ..level_utilized.memory import save_temporary_memory, get_temporary_memory

from ..level_utilized.utility import agent_creator, summarize_system_prompt, summarize_message_prompt

from ...client.tasks.tasks import Task
from ...client.tasks.task_response import ObjectResponse

from ..level_one.call import Call


class AgentManager:
    async def agent(
        self,
        agent_id: str,
        prompt: str,
        response_format: BaseModel = str,
        tools: list[str] = [],
        context: Any = None,
        llm_model: str = "openai/gpt-4o",
        system_prompt: Optional[Any] = None,
        retries: int = 1,
        context_compress: bool = False,
        memory: bool = False
    ):
        try:
            roulette_agent = agent_creator(
                response_format=response_format, 
                tools=tools, 
                context=context, 
                llm_model=llm_model, 
                system_prompt=system_prompt,
                context_compress=context_compress
            )

            roulette_agent.retries = retries
            
            message_history = None
            if memory:
                message_history = get_temporary_memory(agent_id)

            message = [{
                "type": "text",
                "text": f"{prompt}"
            }]

            if "claude-3-5-sonnet" in llm_model:
                print("Tools", tools)
                if "ComputerUse.*" in tools:
                    try:
                        from ..level_utilized.cu import ComputerUse_screenshot_tool
                        result_of_screenshot = ComputerUse_screenshot_tool()
                        message.append(result_of_screenshot)
                    except Exception as e:
                        print("Error", e)

            feedback = ""
            satisfied = False
            total_request_tokens = 0
            total_response_tokens = 0
            total_retries = 0

            while not satisfied:
                message[0]["text"] = message[0]["text"] + "\n\n" + feedback
                print("message: ", message)

                try:
                    print("I sent the request3")
                    result = await roulette_agent.run(message, message_history=message_history)
                    print("I got the response3")
                except (openai.BadRequestError, anthropic.BadRequestError) as e:
                    str_e = str(e)
                    if "400" in str_e and context_compress:
                        try:
                            # These functions are not async, so don't await them
                            compressed_prompt = summarize_system_prompt(system_prompt, llm_model)
                            if compressed_prompt:
                                print("compressed_prompt", compressed_prompt)
                            message[0]["text"] = summarize_message_prompt(message[0]["text"], llm_model)
                            if message[0]["text"]:
                                print("compressed_message", message[0]["text"])

                            roulette_agent = agent_creator(
                                response_format=response_format,
                                tools=tools,
                                context=context,
                                llm_model=llm_model,
                                system_prompt=compressed_prompt,
                                context_compress=False
                            )
                            print("I sent the request4")
                            result = await roulette_agent.run(message, message_history=message_history)
                            print("I got the response4")
                        except Exception as e:
                            traceback.print_exc()
                            return {"status_code": 403, "detail": "Error processing Agent request: " + str(e)}
                    else:
                        return {"status_code": 403, "detail": "Error processing Agent request: " + str(e)}

                total_request_tokens += result.usage().request_tokens
                total_response_tokens += result.usage().response_tokens

                if retries == 1:
                    satisfied = True
                elif total_retries >= retries:
                    satisfied = True
                else:
                    total_retries += 1
                    print("Retrying", total_retries)

                    try:
                        class Satisfying(ObjectResponse):
                            satisfied: bool
                            feedback: str
                            
                        from ...client.level_two.agent import OtherTask
                        other_task = OtherTask(task=prompt, result=result.data)

                        satify_result = await Call.gpt_4o(
                            "Check if the result is satisfied", 
                            response_format=Satisfying, 
                            context=other_task, 
                            llm_model=llm_model
                        )
                        feedback = satify_result["result"].feedback

                        satisfied = satify_result["result"].satisfied
                    except Exception as e:
                        traceback.print_exc()
                        satisfied = True  # Break the loop on error

            if memory:
                save_temporary_memory(result.message_history, agent_id)

            return {
                "status_code": 200, 
                "result": result.data, 
                "usage": {
                    "input_tokens": total_request_tokens, 
                    "output_tokens": total_response_tokens
                }
            }

        except Exception as e:
            traceback.print_exc()
            return {"status_code": 500, "detail": f"Error processing Agent request: {str(e)}"}


Agent = AgentManager()


import base64
import inspect
import subprocess
import traceback
import asyncio
from typing import List, Dict, Any, Optional, Union, Callable
from pydantic import BaseModel



from fastapi import HTTPException
from pydantic import BaseModel
from mcp import ClientSession, StdioServerParameters
from mcp.client.stdio import stdio_client
from mcp.client.stdio import get_default_environment
import asyncio
from contextlib import asynccontextmanager
@asynccontextmanager
async def managed_session(command: str, args: list, env: dict | None = None):

    if env is None:
        env = get_default_environment()
    else:
        default_env = get_default_environment()
        default_env.update(env)
        env = default_env


    server_params = StdioServerParameters(
        command=command,
        args=args,
        env=env,
    )

    
    client = None
    session = None
    
    try:
        client = stdio_client(server_params)

        read, write = await client.__aenter__()
        session = ClientSession(read, write)
        await session.__aenter__()
        await session.initialize()
        yield session
    finally:
        if session:
            try:
                await session.__aexit__(None, None, None)
            except Exception:
                pass
        if client:
            try:
                await client.__aexit__(None, None, None)
            except Exception:
                pass


def install_library_(library):
    try:
        result = subprocess.run(
            ["uv", "pip", "install", library],
            check=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        return result.returncode == 0
    except subprocess.CalledProcessError:

        return False


def uninstall_library_(library):
    try:
        result = subprocess.run(
            ["uv", "pip", "uninstall", "-y", library],
            check=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        return result.returncode == 0
    except subprocess.CalledProcessError:

        return False
    

def add_tool_(function, description: str = "", properties: Dict[str, Any] = None, required: List[str] = None):
    """
    Add a tool to the registered functions.
    
    Args:
        function: The function to be registered as a tool
    """
    from ..server.function_tools import tool
    # Apply the tool decorator with empty description


    
    decorated_function = tool(description=description, custom_properties=properties, custom_required=required)(function)
    return decorated_function

    






import cloudpickle
cloudpickle.DEFAULT_PROTOCOL = 2
from fastapi import HTTPException
from pydantic import BaseModel
from mcp import ClientSession, StdioServerParameters

import asyncio
from contextlib import asynccontextmanager
# Create server parameters for stdio connection

from .api import app, timeout


prefix = "/tools"


class InstallLibraryRequest(BaseModel):
    library: str



@app.post(f"{prefix}/install_library")
@timeout(30.0)
async def install_library(request: InstallLibraryRequest):
    """
    Endpoint to install a library.

    Args:
        library: The library to install

    Returns:
        A success message
    """


    install_library_(request.library)

    return {"message": "Library installed successfully"}



@app.post(f"{prefix}/uninstall_library")
@timeout(30.0)
async def uninstall_library(request: InstallLibraryRequest):
    """
    Endpoint to uninstall a library.
    """
    uninstall_library_(request.library)
    return {"message": "Library uninstalled successfully"}





class AddToolRequest(BaseModel):
    function: str

@app.post(f"{prefix}/add_tool")
@timeout(30.0)
async def add_tool(request: AddToolRequest):
    """
    Endpoint to add a tool.
    """
    # Cloudpickle the function
    decoded_function = base64.b64decode(request.function)
    deserialized_function = cloudpickle.loads(decoded_function)



    add_tool_(deserialized_function)
    return {"message": "Tool added successfully"}



class AddMCPToolRequest(BaseModel):
    name: str
    command: str
    args: List[str]
    env: Dict[str, str]


async def add_mcp_tool_(name: str, command: str, args: List[str], env: Dict[str, str]):
    """
    Add a tool.
    """
    def get_python_type(schema_type: str, format: Optional[str] = None) -> type:
        """Convert JSON schema type to Python type."""
        type_mapping = {
            "string": str,
            "integer": int,
            "boolean": bool,
            "number": float,
            "array": list,
            "object": dict,
        }
        return type_mapping.get(schema_type, Any)

    async with managed_session(command=command, args=args, env=env) as session:
        tools = await session.list_tools()


        tools = tools.tools
        for tool in tools:

            tool_name: str = tool.name
            tool_desc: str = tool.description
            input_schema: Dict[str, Any] = tool.inputSchema
            properties: Dict[str, Dict[str, Any]] = input_schema.get("properties", {})
            required: List[str] = input_schema.get("required", [])

            def create_tool_function(
                tool_name: str,
                properties: Dict[str, Dict[str, Any]],
                required: List[str],
            ) -> Callable[..., Dict[str, Any]]:
                # Create function parameters type annotations
                annotations = {}
                defaults = {}

                # First add required parameters
                for param_name in required:
                    param_info = properties[param_name]
                    param_type = get_python_type(param_info.get("type", "any"))
                    annotations[param_name] = param_type

                # Then add optional parameters
                for param_name, param_info in properties.items():
                    if param_name not in required:
                        param_type = get_python_type(param_info.get("type", "any"))
                        annotations[param_name] = param_type
                        defaults[param_name] = param_info.get("default", None)

                # Create the signature parameters
                from inspect import Parameter, Signature
                
                parameters = []
                # Add required parameters first
                for param_name in required:
                    param_type = annotations[param_name]
                    parameters.append(
                        Parameter(
                            name=param_name,
                            kind=Parameter.POSITIONAL_OR_KEYWORD,
                            annotation=param_type
                        )
                    )
                
                # Add optional parameters
                for param_name, param_type in annotations.items():
                    if param_name not in required:
                        parameters.append(
                            Parameter(
                                name=param_name,
                                kind=Parameter.POSITIONAL_OR_KEYWORD,
                                annotation=param_type,
                                default=defaults[param_name]
                            )
                        )

                async def tool_function(*args: Any, **kwargs: Any) -> Dict[str, Any]:
                    # Convert positional args to kwargs
                    if len(args) > len(required):
                        raise TypeError(
                            f"{tool_name}() takes {len(required)} positional arguments but {len(args)} were given"
                        )

                    # Combine positional args with kwargs
                    all_kwargs = kwargs.copy()
                    for i, arg in enumerate(args):
                        if i < len(required):
                            all_kwargs[required[i]] = arg

                    # Validate required parameters
                    for req in required:
                        if req not in all_kwargs:
                            raise ValueError(f"Missing required parameter: {req}")

                    # Add defaults for optional parameters
                    for param, default in defaults.items():
                        if param not in all_kwargs:
                            all_kwargs[param] = default

                    async with managed_session(command=tool_function.command, args=tool_function.args, env=tool_function.env) as new_session:
                        # Remove None kwargs
                        all_kwargs = {k: v for k, v in all_kwargs.items() if v is not None}
                        result = await new_session.call_tool(name=tool_name, arguments=all_kwargs)
                        return {"result": result}

                # Set function name and annotations
                tool_function.__name__ = tool_name
                tool_function.__annotations__ = {
                    **annotations,
                    "return": Dict[str, Any],
                }
                tool_function.__doc__ = f"{tool_desc}\n\nReturns:\n    Tool execution results"

                # Create and set the signature
                tool_function.__signature__ = Signature(
                    parameters=parameters,
                    return_annotation=Dict[str, Any]
                )

                # Store session parameters as attributes of the function
                tool_function.command = command
                tool_function.args = args
                tool_function.env = env

                return tool_function

            # Create function with proper annotations
            func = create_tool_function(tool_name, properties, required)
            #name should be name__function_name
            full_name = f"{name}__{tool_name}"
            func.__name__ = full_name



            add_tool_(func, description=tool_desc, properties=properties, required=required)


@app.post(f"{prefix}/add_mcp_tool")
@timeout(60.0)
async def add_mcp_tool(request: AddMCPToolRequest):
    """
    Endpoint to add a tool.
    """
    await add_mcp_tool_(request.name, request.command, request.args, request.env)
    return {"message": "Tool added successfully"}


from ..agent_configuration.agent_configuration import get_or_create_client, register_tools
from ..tasks.tasks import Task
from typing import Any


class Direct:
    """A callable class for making direct LLM calls using the Upsonic client."""
    
    @staticmethod
    def do(task: Task, model: str | None = None, client: Any = None, debug: bool = False):
        """
        Execute a direct LLM call with the given task and model.
        
        Args:
            task: The task to execute
            model: The LLM model to use (default: "openai/gpt-4")
            client: Optional custom client to use instead of creating a new one
            
        Returns:
            The response from the LLM
        """
        global latest_upsonic_client
        from ..latest_upsonic_client import latest_upsonic_client

        # Use provided client or get/create one
        if client is not None:
            the_client = client
        else:
            the_client = get_or_create_client(debug=debug)
        
        # Register tools if needed
        the_client = register_tools(the_client, task.tools)

        # Execute the direct call
        return the_client.call(task, model)
    

    @staticmethod
    def print_do(task: Task, model: str | None = None, client: Any = None):
        result = Direct.do(task, model, client)
        print(result)
        return result



"""
Upsonic Tools Module
This module contains the tool implementations that can be used with Upsonic.
"""
from typing import Any, List, Dict, Optional, Type, Union, Callable

class Search:
    pass

class ComputerUse:
    pass

class BrowserUse:
    pass

# Export all tool classes
__all__ = ["Search", "ComputerUse", "BrowserUse"] 


import copy
from datetime import datetime
import dill
import cloudpickle
cloudpickle.DEFAULT_PROTOCOL = 2
import base64

from pydantic import BaseModel
from ..knowledge_base.knowledge_base import KnowledgeBase
from ...exception import (
    NoAPIKeyException,
    ContextWindowTooSmallException,
    InvalidRequestException,
    UnsupportedLLMModelException,
    CallErrorException
)


def serialize_context(context, client):
    if isinstance(context, KnowledgeBase):
        context = context.markdown(client)
    
    return context

def context_serializer(context, client):

    if context is None:
        context = []


    copy_of_context = copy.deepcopy(context)

    if not isinstance(copy_of_context, list):
        copy_of_context = [copy_of_context]
    

    # Adding current date time to the context
    copy_of_context.append(f"Current date and time: {datetime.now().strftime('%Y-%m-%d %H:%M:%S')}")

    for i, each in enumerate(copy_of_context):
            try:
                each.tools = []
            except:
                pass
            try:
                each.response_format = None
            except:
                pass

            copy_of_context[i] = serialize_context(each, client)


    the_module = dill.detect.getmodule(copy_of_context)
    if the_module is not None:
        cloudpickle.register_pickle_by_value(the_module)


    pickled_context = cloudpickle.dumps(copy_of_context)
    context = base64.b64encode(pickled_context).decode("utf-8")


    return context



def response_format_serializer(response_format):
    if response_format is None:
        response_format_str = "str"
    elif isinstance(response_format, (type, BaseModel)):
        # If it's a Pydantic model or other type, cloudpickle and base64 encode it
        the_module = dill.detect.getmodule(response_format)
        if the_module is not None:
            cloudpickle.register_pickle_by_value(the_module)
        pickled_format = cloudpickle.dumps(response_format)
        response_format_str = base64.b64encode(pickled_format).decode("utf-8")
    else:
        response_format_str = "str"

    return response_format_str


def response_format_deserializer(response_format_str, result):
    if response_format_str != "str":
        decoded_result = base64.b64decode(result["result"])
        deserialized_result = cloudpickle.loads(decoded_result)
    else:
        deserialized_result = result["result"]

    result["result"] = deserialized_result

    return result


def tools_serializer(tools_):
    tools = []
    for i in tools_:


        if isinstance(i, type):

            tools.append(i.__name__+".*")
        # If its a string, get the name of the string
        elif isinstance(i, str):

            tools.append(i)
        elif isinstance(i, object):
            sub_i = i.__class__
            tools.append(sub_i.__name__+".*")
    return tools



def error_handler(result):
    if result["status_code"] == 401:
        raise NoAPIKeyException(result["detail"])
    
    if result["status_code"] == 402:
        raise ContextWindowTooSmallException(result["detail"])

    if result["status_code"] == 403:
        raise InvalidRequestException(result["detail"])

    if result["status_code"] == 400:
        raise UnsupportedLLMModelException(result["detail"])

    if result["status_code"] == 500:
        raise CallErrorException(result)


class NoAPIKeyException(Exception):
    """Raised when no API key is provided."""
    pass

class UnsupportedLLMModelException(Exception):
    """Raised when an unsupported LLM model is specified."""
    pass

class ContextWindowTooSmallException(Exception):
    """Raised when the context window is too small for the input."""
    pass

class InvalidRequestException(Exception):
    """Raised when the request is invalid."""
    pass

class CallErrorException(Exception):
    """Raised when there is an error in making a call."""
    pass

class ServerStatusException(Exception):
    """Custom exception for server status check failures."""
    pass

class TimeoutException(Exception):
    """Custom exception for request timeout."""
    pass

class ToolError(Exception):
    """Raised when a tool encounters an error."""
    def __init__(self, message):
        self.message = message



from pydantic import BaseModel
from pydantic_ai.result import ResultDataT_inv, ResultDataT

from typing import Any, Optional

from ...storage.configuration import Configuration

from ..level_utilized.utility import agent_creator, summarize_message_prompt

import openai
import traceback



class CallManager:
    async def gpt_4o(
        self,
        prompt: str,
        response_format: BaseModel = str,
        tools: list[str] = [],
        context: Any = None,
        llm_model: str = "openai/gpt-4o",
        system_prompt: Optional[Any] = None 
    ):
        roulette_agent = agent_creator(response_format, tools, context, llm_model, system_prompt)
        
        message = [{
            "type": "text",
            "text": f"{prompt}"
        }]

        try:
            if "claude-3-5-sonnet" in llm_model:
                print("Tools", tools)
                if "ComputerUse.*" in tools:
                    try:
                        from ..level_utilized.cu import ComputerUse_screenshot_tool
                        result_of_screenshot = ComputerUse_screenshot_tool()
                        message.append(result_of_screenshot)
                    except Exception as e:
                        print("Error", e)

            print("I sent the request1")
            result = await roulette_agent.run(message)
            print("I got the response1")
            usage = result.usage()

            return {"status_code": 200, "result": result.data, "usage": {"input_tokens": usage.request_tokens, "output_tokens": usage.response_tokens}}
        except AttributeError:
            return roulette_agent
        except openai.BadRequestError as e:
            str_e = str(e)
            if "400" in str_e:
                # Try to compress the message prompt - this is not async
                try:
                    message[0]["text"] = summarize_message_prompt(message[0]["text"], llm_model)
                    print("I sent the request2")
                    result = await roulette_agent.run(message)
                    print("I got the response2")
                except Exception as e:
                    traceback.print_exc()
                    return {"status_code": 403, "detail": "Error processing request: " + str(e)}
            else:
                return {"status_code": 403, "detail": "Error processing request: " + str(e)}

            usage = result.usage()
            return {"status_code": 200, "result": result.data, "usage": {"input_tokens": usage.request_tokens, "output_tokens": usage.response_tokens}}

Call = CallManager()



import inspect
import traceback
import types
from itertools import chain
from pydantic_ai import Agent
from pydantic_ai.models.openai import OpenAIModel, OpenAIAgentModel
from pydantic_ai.models.anthropic import AnthropicModel
from openai import AsyncOpenAI, NOT_GIVEN
from openai import AsyncAzureOpenAI
import hashlib

from pydantic import BaseModel
from fastapi import HTTPException, status
from functools import wraps
from typing import Any, Callable, Optional
from pydantic_ai import RunContext, Tool
from anthropic import AsyncAnthropicBedrock
from dataclasses import dataclass
from openai.types.chat import ChatCompletion, ChatCompletionChunk
from openai.types import chat
from collections.abc import AsyncIterator
from typing import Literal
from openai import AsyncStream


from ...storage.configuration import Configuration
from ...storage.caching import save_to_cache_with_expiry, get_from_cache_with_expiry

from ...tools_server.function_client import FunctionToolManager


my_settings_openai = dict(parallel_tool_calls=False)
my_settings_anthropic = dict(parallel_tool_calls=False)



def tool_wrapper(func: Callable) -> Callable:
    @wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        # Log the tool call
        tool_name = getattr(func, "__name__", str(func))

        
        try:
            # Call the original function
            result = func(*args, **kwargs)

            return result
        except Exception as e:
            print("Tool call failed:", e)
            return {"status_code": 500, "detail": f"Tool call failed: {e}"}
    
    return wrapper


def summarize_text(text: str, llm_model: Any, chunk_size: int = 100000, max_size: int = 300000) -> str:
    """Base function to summarize any text by splitting into chunks and summarizing each."""
    # Return early if text is None or empty
    if text is None:
        return ""
    
    if not isinstance(text, str):
        try:
            text = str(text)
        except:
            return ""

    if not text:
        return ""

    # If text is already under max_size, return it
    if len(text) <= max_size:
        return text

    # Generate a cache key based on text content and parameters
    cache_key = hashlib.md5(f"{text}{llm_model}{chunk_size}{max_size}".encode()).hexdigest()
    
    # Try to get from cache first
    cached_result = get_from_cache_with_expiry(cache_key)
    if cached_result is not None:
        print("Using cached summary")
        return cached_result

    # Adjust chunk size based on model
    if "gpt" in str(llm_model).lower():
        # OpenAI has a 1M character limit, we'll use a much smaller chunk size to be safe
        chunk_size = min(chunk_size, 100000)  # 100K per chunk for OpenAI
    elif "claude" in str(llm_model).lower():
        chunk_size = min(chunk_size, 200000)  # 200K per chunk for Claude
    
    try:
        print(f"Original text length: {len(text)}")
        
        # If text is extremely long, do an initial aggressive truncation
        if len(text) > 2000000:  # If over 2M characters
            text = text[:2000000]  # Take first 2M characters
            print("Text was extremely long, truncated to 2M characters")
        
        chunks = [text[i:i + chunk_size] for i in range(0, len(text), chunk_size)]
        print(f"Number of chunks: {len(chunks)}")
        
        model = agent_creator(response_format=str, tools=[], context=None, llm_model=llm_model, system_prompt=None)
        if isinstance(model, dict) and "status_code" in model:
            print(f"Error creating model: {model}")
            return text[:max_size]
        
        # Process chunks in smaller batches if there are too many
        batch_size = 5
        summarized_chunks = []
        
        for batch_start in range(0, len(chunks), batch_size):
            batch_end = min(batch_start + batch_size, len(chunks))
            batch = chunks[batch_start:batch_end]
            
            for i, chunk in enumerate(batch):
                chunk_num = batch_start + i + 1
                try:
                    print(f"Processing chunk {chunk_num}/{len(chunks)}, length: {len(chunk)}")
                    
                    # Create a more focused prompt for better summarization
                    prompt = (
                        "Please provide an extremely concise summary of the following text. "
                        "Focus only on the most important points and key information. "
                        "Be as brief as possible while retaining critical meaning:\n\n"
                    )
                    
                    message = [{"type": "text", "text": prompt + chunk}]
                    result = model.run_sync(message)
                    
                    if result and hasattr(result, 'data') and result.data:
                        # Ensure the summary isn't too long
                        summary = result.data[:max_size//len(chunks)]
                        summarized_chunks.append(summary)
                    else:
                        print(f"Warning: Empty or invalid result for chunk {chunk_num}")
                        # Include a shorter truncated version as fallback
                        summarized_chunks.append(chunk[:500] + "...")
                except Exception as e:
                    print(f"Error summarizing chunk {chunk_num}: {str(e)}")
                    # Include a shorter truncated version as fallback
                    summarized_chunks.append(chunk[:500] + "...")

        # Combine all summarized chunks
        combined_summary = "\n\n".join(summarized_chunks)
        
        # If still too long, recursively summarize with smaller chunks
        if len(combined_summary) > max_size:
            print(f"Combined summary still too long ({len(combined_summary)} chars), recursively summarizing...")
            return summarize_text(
                combined_summary, 
                llm_model, 
                chunk_size=max(5000, chunk_size//4),  # Reduce chunk size more aggressively
                max_size=max_size
            )
            
        print(f"Final summary length: {len(combined_summary)}")
        
        # Cache the result for 1 hour (3600 seconds)
        save_to_cache_with_expiry(combined_summary, cache_key, 3600)
        
        return combined_summary
    except Exception as e:
        traceback.print_exc()
        print(f"Error in summarize_text: {str(e)}")
        # If all else fails, return a truncated version
        return text[:max_size]

def summarize_message_prompt(message_prompt: str, llm_model: Any) -> str:
    """Summarizes the message prompt to reduce its length while preserving key information."""
    print("\n\n\n****************Summarizing message prompt****************\n\n\n")
    if message_prompt is None:
        return ""
    
    try:
        # Use a smaller max size for message prompts
        max_size = 50000  # 100K for messages
        summarized_message_prompt = summarize_text(message_prompt, llm_model, max_size=max_size)
        if summarized_message_prompt is None:
            return ""
        print("Before summarize_message_prompt length: ", len(message_prompt))
        print(f"Summarized message prompt length: {len(summarized_message_prompt)}")
        return summarized_message_prompt
    except Exception as e:
        print(f"Error in summarize_message_prompt: {str(e)}")
        try:
            return str(message_prompt)[:50000] if message_prompt else ""
        except:
            return ""

def summarize_system_prompt(system_prompt: str, llm_model: Any) -> str:
    """Summarizes the system prompt to reduce its length while preserving key information."""
    print("\n\n\n****************Summarizing system prompt****************\n\n\n")
    if system_prompt is None:
        return ""
    
    try:
        # Use a smaller max size for system prompts
        max_size = 50000  # 100K for system prompts
        summarized_system_prompt = summarize_text(system_prompt, llm_model, max_size=max_size)
        if summarized_system_prompt is None:
            return ""
        print("Before summarize_system_prompt length: ", len(system_prompt))
        print(f"Summarized system prompt length: {len(summarized_system_prompt)}")
        return summarized_system_prompt
    except Exception as e:
        print(f"Error in summarize_system_prompt: {str(e)}")
        try:
            return str(system_prompt)[:50000] if system_prompt else ""
        except:
            return ""

def summarize_context_string(context_string: str, llm_model: Any) -> str:
    """Summarizes the context string to reduce its length while preserving key information."""
    print("\n\n\n****************Summarizing context string****************\n\n\n")
    if context_string is None or context_string == "":
        return ""
    
    try:
        # Use a smaller max size for context strings
        max_size = 50000  # 50K for context strings
        summarized_context = summarize_text(context_string, llm_model, max_size=max_size)
        if summarized_context is None:
            return ""
        print("Before summarize_context_string length: ", len(context_string))
        print(f"Summarized context string length: {len(summarized_context)}")
        return summarized_context
    except Exception as e:
        print(f"Error in summarize_context_string: {str(e)}")
        try:
            return str(context_string)[:50000] if context_string else ""
        except:
            return ""

def agent_creator(
        response_format: BaseModel = str,
        tools: list[str] = [],
        context: Any = None,
        llm_model: str = "openai/gpt-4o",
        system_prompt: Optional[Any] = None,
        context_compress: bool = False
    ):

        if llm_model == "openai/gpt-4o" or llm_model == "gpt-4o":
            openai_api_key = Configuration.get("OPENAI_API_KEY")
            if not openai_api_key:
                return {"status_code": 401, "detail": "No API key provided. Please set OPENAI_API_KEY in your configuration."}
            client = AsyncOpenAI(
                api_key=openai_api_key,  # This is the default and can be omitted
            )

            model = OpenAIModel('gpt-4o', openai_client=client,)

        elif llm_model == "openai/o3-mini":
            openai_api_key = Configuration.get("OPENAI_API_KEY")
            if not openai_api_key:
                return {"status_code": 401, "detail": "No API key provided. Please set OPENAI_API_KEY in your configuration."}
            client = AsyncOpenAI(
                api_key=openai_api_key,  # This is the default and can be omitted
            )

            model = OpenAIModel('o3-mini', openai_client=client)

        elif llm_model == "openai/gpt-4o-mini":
            openai_api_key = Configuration.get("OPENAI_API_KEY")
            if not openai_api_key:
                return {"status_code": 401, "detail": "No API key provided. Please set OPENAI_API_KEY in your configuration."}
            client = AsyncOpenAI(
                api_key=openai_api_key,  # This is the default and can be omitted
            )

            model = OpenAIModel('gpt-4o-mini', openai_client=client)



        elif llm_model == "deepseek/deepseek-chat":
            deepseek_api_key = Configuration.get("DEEPSEEK_API_KEY")
            if not deepseek_api_key:
                return {"status_code": 401, "detail": "No API key provided. Please set DEEPSEEK_API_KEY in your configuration."}

            model = OpenAIModel(
                'deepseek-chat',
                base_url='https://api.deepseek.com',
                api_key=deepseek_api_key,
            )




        elif llm_model == "claude/claude-3-5-sonnet" or llm_model == "claude-3-5-sonnet":
            anthropic_api_key = Configuration.get("ANTHROPIC_API_KEY")
            if not anthropic_api_key:
                return {"status_code": 401, "detail": "No API key provided. Please set ANTHROPIC_API_KEY in your configuration."}
            model = AnthropicModel("claude-3-5-sonnet-latest", api_key=anthropic_api_key)




        elif llm_model == "bedrock/claude-3-5-sonnet" or llm_model == "claude-3-5-sonnet-aws":
            aws_access_key_id = Configuration.get("AWS_ACCESS_KEY_ID")
            aws_secret_access_key = Configuration.get("AWS_SECRET_ACCESS_KEY")
            aws_region = Configuration.get("AWS_REGION")

            if not aws_access_key_id or not aws_secret_access_key or not aws_region:
                return {"status_code": 401, "detail": "No AWS credentials provided. Please set AWS_ACCESS_KEY_ID, AWS_SECRET_ACCESS_KEY, and AWS_REGION in your configuration."}
            
            model = AsyncAnthropicBedrock(
                aws_access_key=aws_access_key_id,
                aws_secret_key=aws_secret_access_key,
                aws_region=aws_region
            )

            model = AnthropicModel("us.anthropic.claude-3-5-sonnet-20241022-v2:0", anthropic_client=model)





        elif llm_model == "azure/gpt-4o" or llm_model == "gpt-4o-azure":
            azure_endpoint = Configuration.get("AZURE_OPENAI_ENDPOINT")
            azure_api_version = Configuration.get("AZURE_OPENAI_API_VERSION")
            azure_api_key = Configuration.get("AZURE_OPENAI_API_KEY")

            missing_keys = []
            if not azure_endpoint:
                missing_keys.append("AZURE_OPENAI_ENDPOINT")
            if not azure_api_version:
                missing_keys.append("AZURE_OPENAI_API_VERSION")
            if not azure_api_key:
                missing_keys.append("AZURE_OPENAI_API_KEY")

            if missing_keys:
                return {
                    "status_code": 401,
                    "detail": f"No API key provided. Please set {', '.join(missing_keys)} in your configuration."
                }

            model = AsyncAzureOpenAI(api_version=azure_api_version, azure_endpoint=azure_endpoint, api_key=azure_api_key)
            model = OpenAIModel('gpt-4o', openai_client=model)

        else:
            return {"status_code": 400, "detail": f"Unsupported LLM model: {llm_model}"}

        context_string = ""
        if context is not None:
            if not isinstance(context, list):
                context = [context]

            if isinstance(context, list):
                for each in context:
                    from ...client.level_two.agent import Characterization
                    from ...client.level_two.agent import OtherTask
                    from ...client.tasks.tasks import Task
                    from ...client.tasks.task_response import ObjectResponse
                    from ...client.knowledge_base.knowledge_base import KnowledgeBase
                    type_string = type(each).__name__
                    the_class_string = None
                    try:
                        the_class_string = each.__bases__[0].__name__
                    except:
                        pass
                

                    if type_string == Characterization.__name__:
                        context_string += f"\n\nThis is your character ```character {each.model_dump()}```"
                    elif type_string == OtherTask.__name__:
                        context_string += f"\n\nContexts from question answering: ```question_answering question: {each.task} answer: {each.result}```"
                    elif type_string == Task.__name__:
                        response = None
                        description = each.description
                        try:
                            response = each.response.dict()
                        except:
                            try:
                                response = each.response.model_dump()
                            except:
                                response = each.response
                                
                        context_string += f"\n\nContexts from question answering: ```question_answering question: {description} answer: {response}```   "
                    elif the_class_string == ObjectResponse.__name__:
                        context_string += f"\n\nContexts from object response: ```Requested Output {each.model_fields}```"
                    else:
                        context_string += f"\n\nContexts ```context {each}```"

        # Compress context string if enabled
        if context_compress and context_string:
            context_string = summarize_context_string(context_string, llm_model)

        system_prompt_ = ()

        if system_prompt is not None:
            system_prompt_ = system_prompt + f"The context is: {context_string}"
        elif context_string != "":
            system_prompt_ = f"You are a helpful assistant. User want to add an context to the task. The context is: {context_string}"
        


        the_model_settings = my_settings_openai if tools and llm_model in ["openai/gpt-4o", "azure/gpt-4o", "openai/o3-mini", "openai/gpt-4o-mini"] else None
        if the_model_settings is None:
            the_model_settings = my_settings_anthropic if tools and llm_model in ["claude/claude-3-5-sonnet", "claude-3-5-sonnet"] else None

        roulette_agent = Agent(
            model,
            result_type=response_format,
            retries=5,
            system_prompt=system_prompt_,
            model_settings=the_model_settings
        )



        the_wrapped_tools = []

        with FunctionToolManager() as function_client:
            the_list_of_tools = function_client.get_tools_by_name(tools)

            for each in the_list_of_tools:
                # Wrap the tool with our wrapper
  
    
                wrapped_tool = tool_wrapper(each)

                the_wrapped_tools.append(wrapped_tool)
            



        for each in the_wrapped_tools:
            # İnspect signature of the tool
            signature = inspect.signature(each)
            roulette_agent.tool_plain(each, retries=5)

        # Computer use
        if "claude-3-5-sonnet" in llm_model:
            print("Tools", tools)
            if "ComputerUse.*" in tools:
                try:
                    from .cu import ComputerUse_tools
                    for each in ComputerUse_tools:
                        roulette_agent.tool_plain(each, retries=5)
                except Exception as e:
                    print("Error", e)

        if "BrowserUse.*" in tools:
            try:
                from .bu import BrowserUse_tools
                from .bu.browseruse import LLMManager
                LLMManager.set_model(llm_model)

                for each in BrowserUse_tools:
                    roulette_agent.tool_plain(each, retries=5)
            except Exception as e:
                print("Error", e)

        return roulette_agent




from pydantic import BaseModel


from typing import Any, List, Dict, Optional, Type, Union


from .task_response import CustomTaskResponse, ObjectResponse

class Task(BaseModel):
    description: str
    tools: list[Any] = []
    response_format: Union[Type[CustomTaskResponse], Type[ObjectResponse], None] = None
    _response: Any = None
    context: Any = None
    
    def __init__(self, description: str = None, **data):
        if description is not None:
            data["description"] = description
        super().__init__(**data)

    @property
    def response(self):

        if self._response is None:
            return None

        if type(self._response) == str:
            return self._response



        if self._response._upsonic_response_type == "custom":
            return self._response.output()
        else:
            return self._response






import cloudpickle
cloudpickle.DEFAULT_PROTOCOL = 2
import dill
import base64
import httpx
from typing import Any, List, Dict, Optional, Type, Union
from pydantic import BaseModel
import os
import tempfile


class Markdown:
    def markdown(self, file_path: str) -> str:
        """
        Upload a file and convert it to markdown.

        Args:
            file_path: Path to the file to convert or a URL to download and convert

        Returns:
            The markdown content
        """
        if file_path.startswith("http"):
            # Download file
            response = httpx.get(file_path)
            response.raise_for_status()

            # Save to temporary .html file
            fd, tmp_path = tempfile.mkstemp(suffix=".html")
            os.write(fd, response.content)
            os.close(fd)

            file_path = tmp_path

        if not os.path.exists(file_path):
            raise FileNotFoundError(f"File not found: {file_path}")

        # Read the file and prepare for upload
        with open(file_path, "rb") as f:
            files = {"file": (os.path.basename(file_path), f)}
            response = self.send_request("/markdown/upload", {}, files=files)
            
        if file_path.startswith("/tmp"):
            os.remove(file_path)  # Delete temporary HTML file

        return response.get("markdown")



import copy
import time
import cloudpickle
cloudpickle.DEFAULT_PROTOCOL = 2
import dill
import base64
import httpx
import hashlib
from typing import Any, List, Dict, Optional, Type, Union, Literal
from pydantic import BaseModel

from ..tasks.tasks import Task

from ..printing import agent_end, agent_total_cost, agent_retry



from ..tasks.task_response import ObjectResponse

from ..agent_configuration.agent_configuration import AgentConfiguration

from ..level_utilized.utility import context_serializer


from ..level_utilized.utility import context_serializer, response_format_serializer, tools_serializer, response_format_deserializer, error_handler



from ...storage.caching import save_to_cache_with_expiry, get_from_cache_with_expiry

from ..tools.tools import Search

from ...reliability_processor import ReliabilityProcessor


class SubTask(ObjectResponse):
    description: str
    sources_can_be_used: List[str]
    required_output: str
    tools: List[str]
class SubTaskList(ObjectResponse):
    sub_tasks: List[SubTask]

class AgentMode(ObjectResponse):
    """Mode selection for task decomposition"""
    selected_mode: Literal["level_no_step", "level_one"]




class SearchResult(ObjectResponse):
    any_customers: bool
    products: List[str]
    services: List[str]
    potential_competitors: List[str]
class CompanyObjective(ObjectResponse):
    objective: str
    goals: List[str]
    state: str
class HumanObjective(ObjectResponse):
    job_title: str
    job_description: str
    job_goals: List[str]
    


class Characterization(ObjectResponse):
    website_content: Union[SearchResult, None]
    company_objective: Union[CompanyObjective, None]
    human_objective: Union[HumanObjective, None]
    name_of_the_human_of_tasks: str = None
    contact_of_the_human_of_tasks: str = None


class OtherTask(ObjectResponse):
    task: str
    result: Any



class Agent:


    def agent_(
        self,
        agent_configuration: AgentConfiguration,
        task: Task,
        llm_model: str = None,
    ) -> Any:
        

        
        start_time = time.time()


        results = []

        try:
            if isinstance(task, list):
                for each in task:
                    the_result = self.send_agent_request(agent_configuration, each, llm_model)
                    the_result["time"] = time.time() - start_time
                    results.append(the_result)
                    agent_end(the_result["result"], the_result["llm_model"], the_result["response_format"], start_time, time.time(), the_result["usage"], the_result["tool_count"], the_result["context_count"], self.debug)
            else:
                the_result = self.send_agent_request(agent_configuration, task, llm_model)
                the_result["time"] = time.time() - start_time
                results.append(the_result)
                agent_end(the_result["result"], the_result["llm_model"], the_result["response_format"], start_time, time.time(), the_result["usage"], the_result["tool_count"], the_result["context_count"], self.debug)
        except Exception as e:

            try:
                from ...server import stop_dev_server, stop_main_server, is_tools_server_running, is_main_server_running

                if is_tools_server_running() or is_main_server_running():
                    stop_dev_server()

            except Exception as e:
                pass

            raise e

        end_time = time.time()

        

        return results
        


    def send_agent_request(
        self,
        agent_configuration: AgentConfiguration,
        task: Task,
        llm_model: str = None,
    ) -> Any:
        from ..trace import sentry_sdk
        from ..level_utilized.utility import CallErrorException
        """
        Call GPT-4 with optional tools and MCP servers.

        Args:
            prompt: The input prompt for GPT-4
            response_format: The expected response format (can be a type or Pydantic model)
            tools: Optional list of tool names to use


        Returns:
            The response in the specified format
        """

        if llm_model is None:
            llm_model = self.default_llm_model

        retry_count = 0
        last_error = None


        while retry_count <= agent_configuration.retries:
            try:
                if retry_count > 0:
                    agent_retry(retry_count, agent_configuration.retries)

                tools = tools_serializer(task.tools)

                response_format = task.response_format
                with sentry_sdk.start_transaction(op="task", name="Agent.send_agent_request") as transaction:
                    with sentry_sdk.start_span(op="serialize"):
                        # Serialize the response format if it's a type or BaseModel
                        response_format_str = response_format_serializer(task.response_format)

                        context = context_serializer(task.context, self)

                    with sentry_sdk.start_span(op="prepare_request"):
                        # Prepare the request data
                        data = {
                            "agent_id": agent_configuration.agent_id,
                            "prompt": task.description,
                            "response_format": response_format_str,
                            "tools": tools or [],
                            "context": context,
                            "llm_model": llm_model,
                            "system_prompt": None,
                            "retries": agent_configuration.retries,
                            "context_compress": agent_configuration.context_compress,
                            "memory": agent_configuration.memory
                        }

                    with sentry_sdk.start_span(op="send_request"):
    
                        result = self.send_request("/level_two/agent", data)

                        result = result["result"]

                        error_handler(result)

                    with sentry_sdk.start_span(op="deserialize"):
                        deserialized_result = response_format_deserializer(response_format_str, result)

                # Process result through reliability layer
                processed_result = ReliabilityProcessor.process_result(
                    deserialized_result["result"], 
                    agent_configuration.reliability_layer,
                    task,
                    llm_model
                )
                task._response = processed_result

                response_format_req = None
                if response_format_str == "str":
                    response_format_req = response_format_str
                else:
                    # Class name
                    response_format_req = response_format.__name__
                
                if context is None:
                    context = []

                len_of_context = len(task.context) if task.context is not None else 0

                return {"result": processed_result, "llm_model": llm_model, "response_format": response_format_req, "usage": deserialized_result["usage"], "tool_count": len(tools), "context_count": len_of_context}

            except CallErrorException as e:
                last_error = e
                retry_count += 1
                if retry_count > agent_configuration.retries:
                    raise last_error
                continue



    def create_characterization(self, agent_configuration: AgentConfiguration, llm_model: str = None):
        tools = [Search]
        
        search_result = None
        company_objective_result = None
        human_objective_result = None
        
        search_task = None
        company_objective_task = None

        # Handle website search if URL is provided
        if agent_configuration.company_url:
            search_task = Task(description=f"Make a search for {agent_configuration.company_url}", tools=tools, response_format=SearchResult)
            self.call(search_task, llm_model=llm_model)
            search_result = search_task.response

        # Handle company objective if provided
        if agent_configuration.company_objective:
            context = [search_task] if search_task else None
            company_objective_task = Task(description=f"Generate the company objective for {agent_configuration.company_objective}", 
                                        tools=tools, 
                                        response_format=CompanyObjective,
                                        context=context)
            self.call(company_objective_task, llm_model=llm_model)
            company_objective_result = company_objective_task.response

        # Handle human objective if job title is provided
        if agent_configuration.job_title:
            context = []
            if search_task:
                context.append(search_task)
            if company_objective_task:
                context.append(company_objective_task)
            
            context = context if context else None
            human_objective_task = Task(description=f"Generate the human objective for {agent_configuration.job_title}", 
                                      tools=tools, 
                                      response_format=HumanObjective,
                                      context=context)
            self.call(human_objective_task, llm_model=llm_model)
            human_objective_result = human_objective_task.response

        total_character = Characterization(
            website_content=search_result,
            company_objective=company_objective_result,
            human_objective=human_objective_result,
            name_of_the_human_of_tasks=agent_configuration.name,
            contact_of_the_human_of_tasks=agent_configuration.contact
        )

        return total_character







    def agent(self, agent_configuration: AgentConfiguration, task: Task,  llm_model: str = None):

        if llm_model is None:
            llm_model = agent_configuration.model

        original_task = task




        copy_agent_configuration = copy.deepcopy(agent_configuration)
        copy_agent_configuration_json = copy_agent_configuration.model_dump_json(include={"job_title", "company_url", "company_objective", "name", "contact"})


        
        the_characterization_cache_key = f"characterization_{hashlib.sha256(copy_agent_configuration_json.encode()).hexdigest()}"

        if agent_configuration.caching:
            the_characterization = get_from_cache_with_expiry(the_characterization_cache_key)
            if the_characterization is None:
                the_characterization = self.create_characterization(agent_configuration, llm_model)
                save_to_cache_with_expiry(the_characterization, the_characterization_cache_key, agent_configuration.cache_expiry)
        else:
            the_characterization = self.create_characterization(agent_configuration, llm_model)



        knowledge_base = None

        if agent_configuration.knowledge_base:
            knowledge_base = self.knowledge_base(agent_configuration, llm_model)
            



        
        the_task = task

        is_it_sub_task = False
        shared_context = []

        if agent_configuration.sub_task:
            sub_tasks = self.multiple(task, llm_model)
            is_it_sub_task = True






            the_task = sub_tasks
    



        if not isinstance(the_task, list):
            the_task = [the_task]


        for each in the_task:
            if not isinstance(each.context, list):
                each.context = [each.context]


        last_task = []
        for each in the_task:
            if isinstance(each.context, list):
                last_task.append(each)
        the_task = last_task


        for each in the_task:
            each.context.append(the_characterization)

        # Add knowledge base to the context for each task
        if knowledge_base:
            if isinstance(the_task, list):
                for each in the_task:
                    if each.context:
                        each.context.append(knowledge_base)
                    else:
                        each.context = [knowledge_base]

        if task.context:
            for each in the_task:
                each.context += task.context
                


        if agent_configuration.tools:
            if isinstance(the_task, list):
                for each in the_task:
                    each.tools = agent_configuration.tools

        


        results = []    
        if isinstance(the_task, list):
            for each in the_task:
                if is_it_sub_task:
                    if shared_context:
                        each.context += shared_context



                result = self.agent_(agent_configuration, each, llm_model=llm_model)
                results += result

                if is_it_sub_task:
                    
                    shared_context.append(OtherTask(task=each.description, result=each.response))






        original_task._response = the_task[-1].response


        
        total_time = 0
        for each in results:
            total_time += each["time"]

        total_input_tokens = 0
        total_output_tokens = 0
        for each in results:

            total_input_tokens += each["usage"]["input_tokens"]
            total_output_tokens += each["usage"]["output_tokens"]

        the_llm_model = llm_model
        if the_llm_model is None:
            the_llm_model = self.default_llm_model

        agent_total_cost(total_input_tokens, total_output_tokens, total_time, the_llm_model)

        return original_task.response




    def multiple(self, task: Task, llm_model: str = None):
        # First, determine the mode of operation
        mode_selection_prompt = f"""
You are a Task Analysis AI that helps determine the best mode of task decomposition.

Given task: "{task.description}"

Analyze the task characteristics:

Level No Step (Direct Execution) is suitable for:
- Tasks that can be completed in a single, atomic operation
- Tasks where the output format is simple and well-defined
- Tasks that don't require setup or configuration
- Tasks where AI can directly generate the complete result
- Tasks without dependencies or external integrations
Examples:
- Simple data transformations
- Direct text generation
- Single API call operations
- Basic calculations or conversions

Level One (Basic Decomposition) is suitable for:
- Tasks requiring multiple steps or verifications
- Tasks with clear, linear steps
- Tasks needing external information or resources
- Tasks requiring setup or configuration
- Tasks involving API integrations or data processing
- Tasks that need error handling
- Information retrieval and verification tasks
Examples of Level One Tasks:
- Finding and verifying documentation
- Implementation tasks with clear steps
- Multi-step data processing
- Tasks requiring setup and configuration
- Tasks involving API usage
- Tasks needing error handling
- Tasks that follow a linear sequence of steps

Select the mode based on these characteristics.
Prefer level_no_step when the task can be completed directly without any decomposition.
Use Level One for any task requiring multiple steps or verification.
"""
        mode_selector = Task(
            description=mode_selection_prompt,
            response_format=AgentMode,
            context=[task]
        )
        
        self.call(mode_selector, llm_model)
        
        # If level_no_step is selected, return just the end task
        if mode_selector.response.selected_mode == "level_no_step":
            return [Task(description=task.description, response_format=task.response_format, tools=task.tools)]

        # Generate a list of sub tasks
        prompt = f"""
You are a Task Decomposition AI that helps break down large tasks into smaller, manageable subtasks.

Given task: "{task.description}"
Available tools: {task.tools if task.tools else "No tools available"}

Tool Dependency Guidelines:
- File Operations: Tasks involving file reading, writing, or manipulation require file system tools
- Terminal Operations: Tasks requiring command execution need terminal access tools
- Web Operations: Tasks involving web searches or API calls need web access tools
- System Operations: Tasks involving system configuration or environment setup need system tools

Task Decomposition Rules:
1. Only create subtasks that can be completed with the available tools
2. Skip any operations that would require unavailable tools
3. Each subtask must be achievable with the given tool set
4. If a critical operation cannot be performed due to missing tools, note it in the task description
5. Adapt the approach based on available tools rather than assuming tool availability

General Task Rules:
1. Each subtask should be clear, specific, and actionable
2. Subtasks should be ordered in a logical sequence
3. Each subtask should be necessary for completing the main task
4. Avoid overly broad or vague subtasks
5. Keep subtasks at a similar level of granularity

Tool Availability Impact:
- Without file system tools: Skip file operations
- Without terminal tools: Avoid command execution tasks
- Without web tools: Skip online searches, API calls
- Without system tools: Avoid system configuration tasks
"""
        sub_tasker_context = [task, task.response_format]
        if task.context:
            sub_tasker_context = task.context
        sub_tasker = Task(description=prompt, response_format=SubTaskList, context=sub_tasker_context, tools=task.tools)

        self.call(sub_tasker, llm_model)

        sub_tasks = []

        # Create tasks from subtasks
        for each in sub_tasker.response.sub_tasks:
            new_task = Task(description=each.description + " " + each.required_output + " " + str(each.sources_can_be_used) + " " + str(each.tools) + "Focus to complete the task with right result, Dont ask to human directly do it and give the result.")
            new_task.tools = task.tools
            sub_tasks.append(new_task)

        # Add the final task that will produce the original desired response format
        end_task = Task(description=task.description, response_format=task.response_format)
        sub_tasks.append(end_task)

        return sub_tasks





    def multi_agent(self, agent_configurations: List[AgentConfiguration], tasks: Any, llm_model: str = None):
        agent_tasks = []

        the_agents = {}

        for each in agent_configurations:
            agent_key = each.agent_id[:5] + "_" + each.job_title
            the_agents[agent_key] = each


        the_agents_keys = list(the_agents.keys())



        class TheAgents_(ObjectResponse):
            agents: List[str]


        the_agents_ = TheAgents_(agents=the_agents_keys)


        class SelectedAgent(ObjectResponse):
            selected_agent: str


        if isinstance(tasks, list) != True:
            tasks = [tasks]

        
        for each in tasks:
            is_end = False
            while is_end == False:
                selecting_task  = Task(description="Select an agent for this task", response_format=SelectedAgent, context=[the_agents_, each])

                the_call_llm_model = agent_configurations[0].model
                self.call(selecting_task, the_call_llm_model)

                if selecting_task.response.selected_agent in the_agents:
                    is_end = True



                agent_tasks.append({
                    "agent": the_agents[selecting_task.response.selected_agent],
                    "task": each
                })
                    


        # Store original client
        original_client = self

        for each in agent_tasks:
            # Check if agent has a custom client
            if each["agent"].client is not None:
                # Use agent's custom client for this task
                each["agent"].client.agent(each["agent"], each["task"], llm_model)
            else:
                # Use the default/automatic client
                original_client.agent(each["agent"], each["task"], llm_model)


        return the_agents





from dataclasses import Field
import uuid
from pydantic import BaseModel


from typing import Any, List, Dict, Optional, Type, Union

from ..knowledge_base.knowledge_base import KnowledgeBase
from ..tasks.tasks import Task

from ..latest_upsonic_client import latest_upsonic_client

def register_tools(client, tools):
    """Register tools with the client."""
    if tools is not None:
        for tool in tools:
            # Handle special tool classes from upsonic.client.tools
            if tool.__module__ == 'upsonic.client.tools':

                client.tool()(tool)
                continue
                
            # If tool is a class (not an instance)
            if isinstance(tool, type):
                if hasattr(tool, 'command') and hasattr(tool, 'args'):

                    client.mcp()(tool)
                else:

                    client.tool()(tool)
            else:
                # Get all attributes of the tool instance/object
                tool_attrs = dir(tool)
                
                # Filter out special methods and get only callable attributes
                functions = [attr for attr in tool_attrs 
                           if not attr.startswith('__') and callable(getattr(tool, attr))]
                
                if functions:
                    # If the tool has functions, use the tool() decorator
                    print("INSTANCE TOOL REGISTERED:", tool.__class__.__name__)

                    if not isinstance(tool, object):
                        client.tool()(tool.__class__)
                    else:
                        client.tool()(tool)
                else:
                    # If the tool has no functions, use mcp()
                    print("INSTANCE MCP REGISTERED:", tool.__class__.__name__)
                    client.mcp()(tool.__class__)
    return client


def get_or_create_client(debug: bool = False):
    """Get existing client or create a new one."""
    
    global latest_upsonic_client
    
    if latest_upsonic_client is not None:
        # Check if the existing client's status is False
        if not latest_upsonic_client.status():
            from ..base import UpsonicClient
            new_client = UpsonicClient("localserver", debug=debug)
            latest_upsonic_client = new_client
        return latest_upsonic_client
    
    from ..base import UpsonicClient
    the_client = UpsonicClient("localserver", debug=debug)
    latest_upsonic_client = the_client
    return the_client


def execute_task(agent_config, task: Task, debug: bool = False):
    """Execute a task with the given agent configuration."""
    global latest_upsonic_client
    
    # If agent has a custom client, use it
    if hasattr(agent_config, 'client') and agent_config.client is not None:
        the_client = agent_config.client
    else:
        # Get or create client using existing process
        the_client = get_or_create_client(debug=debug)
    
    # Register tools if needed
    the_client = register_tools(the_client, task.tools)
    
    the_client.run(agent_config, task)
    return task.response


class AgentConfiguration(BaseModel):
    agent_id_: str = None
    job_title: str
    company_url: str = None
    company_objective: str = None
    name: str = ""
    contact: str = ""
    model: str = "openai/gpt-4o"
    client: Any = None  # Add client parameter
    debug: bool = False
    reliability_layer: Any = None  # Changed to Any to accept any class or instance

    def __init__(self, job_title: str = None, client: Any = None, **data):
        if job_title is not None:
            data["job_title"] = job_title
        if client is not None:
            data["client"] = client
        super().__init__(**data)

    sub_task: bool = True
    reflection: bool = True
    memory: bool = False
    caching: bool = True
    cache_expiry: int = 60 * 60
    knowledge_base: KnowledgeBase = None
    tools: List[Any] = []
    context_compress: bool = True

    @property
    def retries(self):
        if self.reflection:
            return 5
        else:
            return 2

    @property
    def agent_id(self):
        if self.agent_id_ is None:
            self.agent_id_ = str(uuid.uuid4())
        return self.agent_id_
    
    def do(self, task: Task):
        return execute_task(self, task, self.debug)
    
    def print_do(self, task: Task):
        result = self.do(task)
        print(result)
        return result



import pytest
import pytest_asyncio
from fastapi.testclient import TestClient
import sys
import os
import asyncio
from fastapi import FastAPI
from httpx import AsyncClient

# Add the src directory to the Python path
sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), "../../src")))

from upsonic.tools_server.server.function_tools import app, _get_json_type

@pytest_asyncio.fixture
async def async_client():
    async with AsyncClient(app=app, base_url="http://test") as client:
        yield client

def test_get_json_type():
    assert _get_json_type(str) == "string"
    assert _get_json_type(int) == "integer"
    assert _get_json_type(bool) == "boolean"
    assert _get_json_type(float) == "number"
    assert _get_json_type(list) == "array"
    assert _get_json_type(dict) == "object"
    assert _get_json_type(object) == "string"  # Default case

@pytest.mark.asyncio
async def test_list_tools(async_client):
    response = await async_client.post("/functions/tools")
    assert response.status_code == 200

    data = response.json()
    assert "available_tools" in data
    tools = data["available_tools"]["tools"]

    # Should have add_numbers and concat_strings
    assert len(tools) == 4

    # Verify add_numbers
    add_numbers = next(t for t in tools if t["name"] == "add_numbers")
    assert add_numbers["description"] == "Add two numbers together"
    assert add_numbers["inputSchema"]["required"] == ["a", "b"]

    # Verify concat_strings
    concat_strings = next(t for t in tools if t["name"] == "concat_strings")
    assert concat_strings["description"] == "Concatenate two strings"
    assert concat_strings["inputSchema"]["required"] == ["str1", "str2"]

@pytest.mark.asyncio
async def test_example_functions(async_client):
    try:
        # Test add_numbers with timeout using asyncio.timeout
        async with asyncio.timeout(5):  # 5 second timeout
            response = await async_client.post(
                "/functions/call_tool",
                json={"tool_name": "add_numbers", "arguments": {"a": 10, "b": 20}},
            )
            assert response.status_code == 200
            assert response.json() == {"result": 30}

        # Test concat_strings
        async with asyncio.timeout(5):
            response = await async_client.post(
                "/functions/call_tool",
                json={"tool_name": "concat_strings", "arguments": {"str1": "Hello, ", "str2": "World!"}},
            )
            assert response.status_code == 200
            assert response.json() == {"result": "Hello, World!"}
    except asyncio.TimeoutError:
        pytest.fail("Test timed out")

@pytest.mark.asyncio
async def test_error_handling(async_client):
    try:
        # Test non-existent tool
        async with asyncio.timeout(5):
            response = await async_client.post(
                "/functions/call_tool", json={"tool_name": "non_existent", "arguments": {}}
            )
            assert response.status_code == 404
            error_response = response.json()
            assert "detail" in error_response
            assert error_response["detail"] == "Tool non_existent not found"

        # Test invalid arguments
        async with asyncio.timeout(5):
            response = await async_client.post(
                "/functions/call_tool",
                json={"tool_name": "add_numbers", "arguments": {"a": "not_a_number", "b": 20}},
            )
            # The API returns a 200 status code with an error message in the response
            assert response.status_code == 200
            error_response = response.json()
            assert "error" in error_response or "detail" in error_response
            # The error message should indicate a failure to call the tool
            error_msg = error_response.get("error", error_response.get("detail", ""))
            assert "failed to call tool" in error_msg.lower() or "can only concatenate" in error_msg.lower()
    except asyncio.TimeoutError:
        pytest.fail("Test timed out")



from ..agent_configuration.agent_configuration import get_or_create_client, register_tools
from ..tasks.tasks import Task
from ..agent_configuration.agent_configuration import AgentConfiguration

from typing import Any

class MultiAgent:
    """A callable class for making direct LLM calls using the Upsonic client."""
    
    @staticmethod
    def do(agents: list[Any], tasks: list[Task], llm_model: str | None = None):
        """
        Execute a direct LLM call with the given task and model.
        
        Args:
            task: The task to execute
            llm_model: The LLM model to use (default: "openai/gpt-4")
            
        Returns:
            The response from the LLM
        """
        global latest_upsonic_client
        from ..latest_upsonic_client import latest_upsonic_client

        # Get or create client for agents without custom clients
        the_client = get_or_create_client()
        
        # Register tools for all tasks regardless of client
        for task in tasks:
            the_client = register_tools(the_client, task.tools)
            # Also register tools for agents with custom clients
            for agent in agents:
                if agent.client is not None:
                    agent.client = register_tools(agent.client, task.tools)
        
        # Update the global client reference
        if latest_upsonic_client is None:
            latest_upsonic_client = the_client

        # Execute the multi-agent call
        return the_client.multi_agent(agents, tasks, llm_model)
    

    @staticmethod
    def print_do(agents: list[Any], tasks: list[Task], llm_model: str | None = None):
        result = MultiAgent.do(agents, tasks, llm_model)
        print(result)
        return result



import cloudpickle
cloudpickle.DEFAULT_PROTOCOL = 2
import dill
import base64
import httpx
import os
from typing import Any, List, Dict, Optional, Type, Union
from pydantic import BaseModel, Field


from dotenv import load_dotenv
load_dotenv(os.path.join(os.getcwd(), ".env"))


class ClientConfig(BaseModel):
    DEFAULT_LLM_MODEL: str = Field(default="openai/gpt-4o")
    
    OPENAI_API_KEY: str | None = Field(default_factory=lambda: os.getenv("OPENAI_API_KEY"))

    ANTHROPIC_API_KEY: str | None = Field(default_factory=lambda: os.getenv("ANTHROPIC_API_KEY"))
    
    AZURE_OPENAI_ENDPOINT: str | None = Field(default_factory=lambda: os.getenv("AZURE_OPENAI_ENDPOINT"))
    AZURE_OPENAI_API_VERSION: str | None = Field(default_factory=lambda: os.getenv("AZURE_OPENAI_API_VERSION"))
    AZURE_OPENAI_API_KEY: str | None = Field(default_factory=lambda: os.getenv("AZURE_OPENAI_API_KEY"))
    
    AWS_ACCESS_KEY_ID: str | None = Field(default_factory=lambda: os.getenv("AWS_ACCESS_KEY_ID"))
    AWS_SECRET_ACCESS_KEY: str | None = Field(default_factory=lambda: os.getenv("AWS_SECRET_ACCESS_KEY"))
    AWS_REGION: str | None = Field(default_factory=lambda: os.getenv("AWS_REGION"))

    DEEPSEEK_API_KEY: str | None = Field(default_factory=lambda: os.getenv("DEEPSEEK_API_KEY"))



class Storage:



    def get_config(self, key: str) -> Any:
        """
        Get a configuration value by key from the server.

        Args:
            key: The configuration key

        Returns:
            The configuration value
        """
        from ..trace import sentry_sdk
        with sentry_sdk.start_transaction(op="task", name="Storage.get_config") as transaction:
            with sentry_sdk.start_span(op="send_request"):
                data = {"key": key}
                response = self.send_request("/storage/config/get", data=data)
            return response.get("value")

    def set_config(self, key: str, value: str) -> str:
        """
        Set a configuration value on the server.

        Args:
            key: The configuration key
            value: The configuration value

        Returns:
            A success message
        """
        from ..trace import sentry_sdk
        with sentry_sdk.start_transaction(op="task", name="Storage.set_config") as transaction:
            with sentry_sdk.start_span(op="send_request"):
                data = {"key": key, "value": value}
                response = self.send_request("/storage/config/set", data=data)
            return response.get("message")


    def set_default_llm_model(self, llm_model: str):
        self.default_llm_model = llm_model


    def config(self, config: ClientConfig):
        # Parse it and use set_config to set it 
        # Exclude the default_llm_model and None values
        for key, value in config.model_dump().items():
            if key != "default_llm_model" and value is not None:
                self.set_config(key, value)
        
        self.default_llm_model = config.DEFAULT_LLM_MODEL



import httpx
from typing import Dict, List, Any, Callable, Optional
from functools import wraps
import inspect


class FunctionToolManager:
    """Client for interacting with the Upsonic Functions API."""

    def __init__(self):
        """Initialize the Upsonic Function client."""
        self.base_url = "http://localhost:8086"

    def get_tools_by_name(self, name: list[str]):
        """
        Get tools by name, supporting wildcard patterns.
        
        Args:
            name: List of tool names or patterns (e.g. ["FileSystem.*", "MyTools.*"])
            
        Returns:
            List of matching tools
        """
        matching_tools = []
        for tool in self.tools():
            tool_name = tool.__name__
            for pattern in name:
                # Handle wildcard pattern
                if pattern.endswith(".*"):
                    prefix = pattern[:-2]  # Remove .* from the end
                    if tool_name.startswith(prefix):
                        matching_tools.append(tool)
                        break
                # Exact match
                elif tool_name == pattern:
                    matching_tools.append(tool)
                    break
        return matching_tools

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    def close(self):
        """Close the client session."""
        pass

    def list_tools(self) -> Dict[str, Any]:
        """List all available tools."""
        with httpx.Client(timeout=600.0) as session:
            response = session.post(f"{self.base_url}/functions/tools")
            response.raise_for_status()
            return response.json()

    def call_tool(self, tool_name: str, arguments: Dict[str, Any]) -> Dict[str, Any]:
        """
        Call a specific tool with the given arguments.

        Args:
            tool_name: Name of the tool to call
            arguments: Dictionary of arguments to pass to the tool

        Returns:
            Tool execution results
        """
        with httpx.Client(timeout=600.0) as session:
            response = session.post(
                f"{self.base_url}/functions/call_tool",
                json={"tool_name": tool_name, "arguments": arguments},
            )
            response.raise_for_status()
            return response.json()

    def tools(self) -> List[Callable[..., Dict[str, Any]]]:
        """Initialize tool-specific methods based on available tools."""
        tools_response = self.list_tools()



        tools = tools_response.get("available_tools", {}).get("tools", [])

        functions: List[Callable[..., Dict[str, Any]]] = []

        def get_python_type(schema_type: str, format: Optional[str] = None) -> type:
            """Convert JSON schema type to Python type."""
            type_mapping = {
                "string": str,
                "integer": int,
                "boolean": bool,
                "number": float,
                "array": list,
                "object": dict,
            }
            return type_mapping.get(schema_type, Any)

        for tool in tools:
            tool_name: str = tool["name"]
            tool_desc: str = tool.get("description", "")
            input_schema: Dict[str, Any] = tool.get("inputSchema", {})
            properties: Dict[str, Dict[str, Any]] = input_schema.get("properties", {})
            required: List[str] = input_schema.get("required", [])

            def create_tool_function(
                tool_name: str,
                properties: Dict[str, Dict[str, Any]],
                required: List[str],
            ) -> Callable[..., Dict[str, Any]]:
                annotations = {}
                defaults = {}
                parameters = []

                # Build parameters for both required and optional arguments
                for param_name in required:
                    param_info = properties[param_name]
                    param_type = get_python_type(param_info.get("type", "any"))
                    annotations[param_name] = param_type
                    parameters.append(
                        inspect.Parameter(
                            param_name,
                            inspect.Parameter.POSITIONAL_OR_KEYWORD,
                            annotation=param_type
                        )
                    )

                for param_name, param_info in properties.items():
                    if param_name not in required:
                        param_type = get_python_type(param_info.get("type", "any"))
                        annotations[param_name] = param_type
                        default_value = param_info.get("default", None)
                        defaults[param_name] = default_value
                        parameters.append(
                            inspect.Parameter(
                                param_name,
                                inspect.Parameter.POSITIONAL_OR_KEYWORD,
                                default=default_value,
                                annotation=param_type
                            )
                        )

                def tool_function(*args: Any, **kwargs: Any) -> Dict[str, Any]:
                    all_kwargs = kwargs.copy()
                    for i, arg in enumerate(args):
                        if i < len(required):
                            all_kwargs[required[i]] = arg

                    for param, default in defaults.items():
                        if param not in all_kwargs:
                            all_kwargs[param] = default

                    return self.call_tool(tool_name, all_kwargs)

                # Create a signature object and apply it to the function
                sig = inspect.Signature(parameters=parameters, return_annotation=Dict[str, Any])
                tool_function.__signature__ = sig
                tool_function.__name__ = tool_name
                tool_function.__annotations__ = {
                    **annotations,
                    "return": Dict[str, Any],
                }
                tool_function.__doc__ = f"{tool_desc}\n\nReturns:\n    Tool execution results"

                return tool_function


            func = create_tool_function(tool_name, properties, required)
            functions.append(func)

        return functions



import asyncio
import base64
import math
import os
import platform
import shlex
import shutil
import tempfile
import time
from enum import StrEnum
from pathlib import Path
from typing import Literal, TypedDict
from uuid import uuid4

# Add import for PyAutoGUI
import pyautogui
from anthropic.types.beta import BetaToolComputerUse20241022Param

from .base import BaseAnthropicTool
from .run import run

OUTPUT_DIR = "/tmp/outputs"

TYPING_DELAY_MS = 12
TYPING_GROUP_SIZE = 50

Action = Literal[
    "key",
    "type",
    "mouse_move",
    "left_click",
    "left_click_drag",
    "right_click",
    "middle_click",
    "double_click",
    "screenshot",
    "cursor_position",
]


class Resolution(TypedDict):
    width: int
    height: int


class ScalingMode(StrEnum):
    AUTO = "auto"  # Automatically determine best scaling
    FIXED = "fixed"  # Use fixed target resolutions
    RELATIVE = "relative"  # Scale by percentage
    NONE = "none"  # No scaling


# Base resolutions for different device categories
DEVICE_CATEGORIES: dict[str, Resolution] = {
    "HD": Resolution(width=1280, height=720),      # 720p
    "FHD": Resolution(width=1920, height=1080),    # 1080p
    "QHD": Resolution(width=2560, height=1440),    # 1440p
    "4K": Resolution(width=3840, height=2160),     # 4K
    "5K": Resolution(width=5120, height=2880),     # 5K (Common in Mac)
    "6K": Resolution(width=6016, height=3384),     # 6K (Pro Display XDR)
}

class ScalingConfig(TypedDict):
    mode: ScalingMode
    target_resolution: Resolution | None  # Used for FIXED mode
    scale_factor: float | None  # Used for RELATIVE mode
    min_scale: float  # Minimum scaling factor
    max_scale: float  # Maximum scaling factor
    preserve_aspect_ratio: bool


DEFAULT_SCALING_CONFIG = ScalingConfig(
    mode=ScalingMode.AUTO,
    target_resolution=None,
    scale_factor=None,
    min_scale=0.25,  # Don't scale below 25%
    max_scale=1.0,   # Don't upscale
    preserve_aspect_ratio=True
)

class ScalingSource(StrEnum):
    COMPUTER = "computer"
    API = "api"


class ComputerToolOptions(TypedDict):
    display_height_px: int
    display_width_px: int
    display_number: int | None


def chunks(s: str, chunk_size: int) -> list[str]:
    return [s[i : i + chunk_size] for i in range(0, len(s), chunk_size)]


def smooth_move_to(x, y, duration=1.2):
    start_x, start_y = pyautogui.position()
    dx = x - start_x
    dy = y - start_y
    distance = math.hypot(dx, dy)  # Calculate the distance in pixels

    start_time = time.time()

    while True:
        elapsed_time = time.time() - start_time
        if elapsed_time > duration:
            break

        t = elapsed_time / duration
        eased_t = (1 - math.cos(t * math.pi)) / 2  # easeInOutSine function

        target_x = start_x + dx * eased_t
        target_y = start_y + dy * eased_t
        pyautogui.moveTo(target_x, target_y)

    # Ensure the mouse ends up exactly at the target (x, y)
    pyautogui.moveTo(x, y)


class ComputerTool(BaseAnthropicTool):
    """
    A tool that allows the agent to interact with the primary monitor's screen, keyboard, and mouse.
    The tool parameters are defined by Anthropic and are not editable.
    """

    name: Literal["computer"] = "computer"
    api_type: Literal["computer_20241022"] = "computer_20241022"
    width: int
    height: int
    display_num: None
    scaling_config: ScalingConfig

    _screenshot_delay = 2.0
    _scaling_enabled = True

    @property
    def options(self) -> ComputerToolOptions:
        width, height = self.scale_coordinates(
            ScalingSource.COMPUTER, self.width, self.height
        )
        return {
            "display_width_px": width,
            "display_height_px": height,
            "display_number": self.display_num,
        }

    def to_params(self) -> BetaToolComputerUse20241022Param:
        return {"name": self.name, "type": self.api_type, **self.options}

    def __init__(self):
        super().__init__()
        self.width, self.height = pyautogui.size()
        self.display_num = None
        self.scaling_config = self._determine_optimal_scaling()

    def _determine_optimal_scaling(self) -> ScalingConfig:
        """Determine the optimal scaling configuration based on the current display."""
        config = DEFAULT_SCALING_CONFIG.copy()
        
        # Get system info
        system = platform.system().lower()
        total_pixels = self.width * self.height
        aspect_ratio = self.width / self.height
        
        # Detect Retina/HiDPI displays on macOS
        is_retina = False
        if system == "darwin":
            try:
                import subprocess
                result = subprocess.run(['system_profiler', 'SPDisplaysDataType'], capture_output=True, text=True)
                is_retina = "Retina" in result.stdout
            except Exception:
                pass

        # Find the closest standard resolution
        closest_category = None
        min_diff = float('inf')
        
        for category, resolution in DEVICE_CATEGORIES.items():
            cat_pixels = resolution["width"] * resolution["height"]
            diff = abs(cat_pixels - total_pixels)
            if diff < min_diff:
                min_diff = diff
                closest_category = category

        # Determine scaling mode and factors
        if is_retina:
            # For Retina displays, we use relative scaling
            config["mode"] = ScalingMode.RELATIVE
            config["scale_factor"] = 0.5  # Default for Retina
        elif total_pixels > DEVICE_CATEGORIES["FHD"]["width"] * DEVICE_CATEGORIES["FHD"]["height"]:
            # For high-res displays, scale down to FHD
            config["mode"] = ScalingMode.FIXED
            config["target_resolution"] = DEVICE_CATEGORIES["FHD"]
        else:
            # For standard/lower resolutions, use minimal scaling
            config["mode"] = ScalingMode.RELATIVE
            config["scale_factor"] = 1.0

        return config

    def scale_coordinates(self, source: ScalingSource, x: int, y: int) -> tuple[int, int]:
        """Scale coordinates based on the current scaling configuration."""
        if not self._scaling_enabled:
            return x, y

        if self.scaling_config["mode"] == ScalingMode.NONE:
            return x, y

        # Get current scaling factors
        if self.scaling_config["mode"] == ScalingMode.FIXED and self.scaling_config["target_resolution"]:
            x_factor = self.scaling_config["target_resolution"]["width"] / self.width
            y_factor = self.scaling_config["target_resolution"]["height"] / self.height
            
            if self.scaling_config["preserve_aspect_ratio"]:
                # Use the same factor for both dimensions to preserve aspect ratio
                x_factor = y_factor = min(x_factor, y_factor)
        
        elif self.scaling_config["mode"] == ScalingMode.RELATIVE and self.scaling_config["scale_factor"]:
            x_factor = y_factor = self.scaling_config["scale_factor"]
        
        else:  # AUTO mode
            # Calculate dynamic scaling factor based on resolution
            base_factor = min(1.0, 1920 / max(self.width, self.height))  # Use FHD as reference
            x_factor = y_factor = max(
                self.scaling_config["min_scale"],
                min(self.scaling_config["max_scale"], base_factor)
            )

        # Apply scaling based on source
        if source == ScalingSource.API:
            # Scale up (from scaled coordinates to actual screen coordinates)
            return round(x / x_factor), round(y / y_factor)
        else:
            # Scale down (from actual screen coordinates to scaled coordinates)
            return round(x * x_factor), round(y * y_factor)

    def update_scaling_config(self, new_config: dict) -> None:
        """Update the scaling configuration with new settings."""
        self.scaling_config.update(new_config)

    async def __call__(
        self,
        *,
        action: Action,
        text: str | None = None,
        coordinate: tuple[int, int] | None = None,
        **kwargs,
    ):
        print("action", action)
        print("text", text)
        print("coordinate", coordinate)
        if action in ("mouse_move", "left_click_drag"):
            if coordinate is None:
                return {"text": f"coordinate is required for {action}"}
            x, y = self.scale_coordinates(
                ScalingSource.API, coordinate[0], coordinate[1]
            )

            if action == "mouse_move":
                smooth_move_to(x, y)
                return {"text": f"Mouse moved to X={x}, Y={y}"}
            elif action == "left_click_drag":
                smooth_move_to(x, y)
                pyautogui.dragTo(x, y, button="left")
                return {"text": f"Mouse dragged to X={x}, Y={y}"}

        elif action in ("key", "type"):
            if text is None:
                return {"text": f"text is required for {action}"}

            if action == "key":
                if platform.system() == "Darwin":  # Check if we're on macOS
                    text = text.replace("super+", "command+")

                # Normalize key names
                def normalize_key(key):
                    key = key.lower().replace("_", "")
                    key_map = {
                        "pagedown": "pgdn",
                        "pageup": "pgup",
                        "enter": "return",
                        "return": "enter",
                    }
                    return key_map.get(key, key)

                keys = [normalize_key(k) for k in text.split("+")]

                if len(keys) > 1:
                    if "darwin" in platform.system().lower():
                        # Use AppleScript for hotkey on macOS
                        keystroke, modifier = (keys[-1], "+".join(keys[:-1]))
                        modifier = modifier.lower() + " down"
                        if keystroke.lower() == "space":
                            keystroke = " "
                        elif keystroke.lower() == "enter":
                            keystroke = "\n"
                        script = f"""
                        tell application "System Events"
                            keystroke "{keystroke}" using {modifier}
                        end tell
                        """
                        os.system("osascript -e '{}'".format(script))
                    else:
                        pyautogui.hotkey(*keys)
                else:
                    pyautogui.press(keys[0])
                return {"text": f"Key pressed: {text}"}
            elif action == "type":
                pyautogui.write(text, interval=TYPING_DELAY_MS / 1000)
                return {"text": f"Text typed: {text}"}

        elif action in ("left_click", "right_click", "double_click", "middle_click"):
            time.sleep(0.1)
            button = {
                "left_click": "left",
                "right_click": "right",
                "middle_click": "middle",
            }
            if action == "double_click":
                pyautogui.click()
                time.sleep(0.1)
                pyautogui.click()
                return {"text": "Double click performed"}
            else:
                pyautogui.click(button=button.get(action, "left"))
                return {"text": f"{action.replace('_', ' ').title()} performed"}

        elif action == "screenshot":
            screenshot_result = self.screenshot()
            return {"type": "image", "source": screenshot_result["source"]}

        elif action == "cursor_position":
            x, y = pyautogui.position()
            x, y = self.scale_coordinates(ScalingSource.COMPUTER, x, y)
            return {"text": f"X={x},Y={y}"}

        else:
            return {"text": f"Invalid action: {action}"}

        # Take a screenshot after the action (except for cursor_position)
        if action != "cursor_position":
            screenshot_result = self.screenshot()
            return {
                "type": "image",
                "text": f"Action '{action}' completed",
                "source": screenshot_result["source"]
            }

    def screenshot(self):
        """Take a screenshot of the current screen and return the base64 encoded image."""
        temp_dir = Path(tempfile.gettempdir())
        path = temp_dir / f"screenshot_{uuid4().hex}.png"

        screenshot = pyautogui.screenshot()
        
        # print current file size before optimization
        screenshot.save(str(path))
        print(f"Original file size: {os.path.getsize(path)} bytes")

        if self._scaling_enabled:
            x, y = self.scale_coordinates(
                ScalingSource.COMPUTER, self.width, self.height
            )
            from PIL import Image

            with Image.open(path) as img:
                # Resize with high-quality downsampling
                img = img.resize((x, y), Image.Resampling.LANCZOS)
                # Save with optimization and reduced quality
                img.save(path)

        if path.exists():
            print(f"Optimized file size: {os.path.getsize(path)} bytes")
            base64_image = base64.b64encode(path.read_bytes()).decode()
            path.unlink()  # Remove the temporary file

            return {
                "type": "image",
                "source": {
                    "type": "base64",
                    "media_type": "image/png",
                    "data": base64_image,
                }
            }
                
        return {"text": "Failed to take screenshot"}

    async def shell(self, command: str, take_screenshot=True):
        """Run a shell command and return the output, error, and optionally a screenshot."""
        _, stdout, stderr = await run(command)
        result = {"text": stdout}
        if stderr:
            result["text"] += f"\nError: {stderr}"

        if take_screenshot:
            # delay to let things settle before taking a screenshot
            await asyncio.sleep(self._screenshot_delay)
            screenshot_result = await self.screenshot()
            result = {
                "type": "image",
                "text": result["text"],
                "source": screenshot_result["source"]
            }

        return result


async def ComputerUse__type(text: str):
    """Execute a typing action using the ComputerTool."""
    tool = ComputerTool()
    return await tool(action="type", text=text)

async def ComputerUse__key(text: str):
    """Execute a key press action using the ComputerTool."""
    tool = ComputerTool()
    return await tool(action="key", text=text)

async def ComputerUse__mouse_move(coordinate: tuple[int, int]):
    """Execute a mouse move action using the ComputerTool."""
    tool = ComputerTool()
    return await tool(action="mouse_move", coordinate=coordinate)

async def ComputerUse__left_click():
    """Execute a left click action using the ComputerTool."""
    tool = ComputerTool()
    return await tool(action="left_click")

async def ComputerUse__right_click():
    """Execute a right click action using the ComputerTool."""
    tool = ComputerTool()
    return await tool(action="right_click")

async def ComputerUse__middle_click():
    """Execute a middle click action using the ComputerTool."""
    tool = ComputerTool()
    return await tool(action="middle_click")

async def ComputerUse__double_click():
    """Execute a double click action using the ComputerTool."""
    tool = ComputerTool()
    return await tool(action="double_click")

async def ComputerUse__left_click_drag(coordinate: tuple[int, int]):
    """Execute a left click drag action using the ComputerTool."""
    tool = ComputerTool()
    return await tool(action="left_click_drag", coordinate=coordinate)

async def ComputerUse__screenshot():
    """Take a screenshot using the ComputerTool."""
    tool = ComputerTool()
    return await tool(action="screenshot")

def ComputerUse_screenshot_tool():
    """Take a screenshot using the ComputerTool and return the base64 encoded image."""
    tool = ComputerTool()
    return tool.screenshot()

async def ComputerUse__cursor_position():
    """Get the current cursor position using the ComputerTool."""
    tool = ComputerTool()
    return await tool(action="cursor_position")

# List of all computer use tools
ComputerUse_tools = [
    ComputerUse__type,
    ComputerUse__key,
    ComputerUse__mouse_move,
    ComputerUse__left_click,
    ComputerUse__right_click,
    ComputerUse__middle_click,
    ComputerUse__double_click,
    ComputerUse__left_click_drag,
    ComputerUse__screenshot,
    ComputerUse__cursor_position
]




import traceback
from fastapi import HTTPException
from pydantic import BaseModel
import inspect
from typing import Any, Dict, List, Type, Callable
from functools import wraps

from .api import app, timeout

prefix = "/functions"

# Registry to store decorated functions
registered_functions: Dict[str, Dict[str, Any]] = {}


def _get_json_type(python_type: Type) -> str:
    """Convert Python type to JSON schema type."""
    type_mapping = {
        str: "string",
        int: "integer",
        bool: "boolean",
        float: "number",
        list: "array",
        dict: "object",
    }
    return type_mapping.get(python_type, "string")


def tool(description: str = "", custom_properties: Dict[str, Any] = None, custom_required: List[str] = None):
    """
    Decorator to register a function as a tool.

    Args:
        description: Optional description of the tool. If not provided, function's docstring will be used.
    """

    def decorator(func: Callable):
        sig = inspect.signature(func)


        # Get parameter info
        properties = {}
        required = []

        # Extract description from docstring if not provided
        tool_description = description
        if not tool_description and func.__doc__:
            # Get the first line of the docstring as description
            tool_description = func.__doc__.strip().split('\n')[0].strip()


        
        for param_name, param in sig.parameters.items():
            param_type = (
                param.annotation if param.annotation != inspect.Parameter.empty else Any
            )
            param_default = (
                None if param.default == inspect.Parameter.empty else param.default
            )

            properties[param_name] = {
                "type": _get_json_type(param_type),
                "description": f"Parameter {param_name}",
            }

            if param_default is not None:
                properties[param_name]["default"] = param_default
            
            # If parameter has no default value, it's required
            if param.default == inspect.Parameter.empty:
                required.append(param_name)


        if custom_properties is not None:
            properties = custom_properties

        if custom_required is not None:
            required = custom_required

        # Register the function with the extracted description
        registered_functions[func.__name__] = {
            "function": func,
            "description": tool_description,
            "properties": properties,
            "required": required,
        }

        # Check if the function is async
        is_async = inspect.iscoroutinefunction(func)

        if is_async:
            @wraps(func)
            async def wrapper(*args, **kwargs):
                return await func(*args, **kwargs)
        else:
            @wraps(func)
            def wrapper(*args, **kwargs):
                return func(*args, **kwargs)

        return wrapper

    return decorator


class ToolRequest(BaseModel):
    tool_name: str
    arguments: dict


@app.post(f"{prefix}/tools")
@timeout(30.0)
async def list_tools():


    tools = []
    for name, info in registered_functions.items():

        tools.append(
            {
                "name": name,
                "description": info["description"],
                "inputSchema": {
                    "type": "object",
                    "properties": info["properties"],
                    "required": info["required"],
                },
            }
        )


    print(tools)    
    return {"available_tools": {"tools": tools}}


@app.post(f"{prefix}/call_tool")
@timeout(30.0)
async def call_tool(request: ToolRequest):


    if request.tool_name not in registered_functions:
        raise HTTPException(
            status_code=404, detail=f"Tool {request.tool_name} not found"
        )

    try:
        func = registered_functions[request.tool_name]["function"]
        # Check if the function is async
        is_async = inspect.iscoroutinefunction(func)
        
        if is_async:
            result = await func(**request.arguments)
        else:
            result = func(**request.arguments)
            

        return {"result": result}
    except Exception as e:

        return {"status_code": 500, "detail": f"Failed to call tool: {str(e)}"}


# Example decorated functions
@tool()
async def add_numbers(a: int, b: int, c: int=0) -> int:
    "Add two numbers together"
    return a + b + c


@tool()
def concat_strings(str1: str, str2: str) -> str:
    "Concatenate two strings"
    return str1 + str2


@tool()
def Search__duckduckgo(query: str, max_results: int = 20) -> list:
    """
    Search the query on DuckDuckGo and return the results.
    """
    try:
        from duckduckgo_search import DDGS

        return list(DDGS().text(query, max_results=max_results))
    except:
        return "An exception occurred"
    

import re
import requests
from bs4 import BeautifulSoup
from urllib.parse import urljoin
@tool()
def Search__read_website(url: str, max_content_length: int = 5000) -> dict:
    """
    Read the content of a website and return the title, meta data, content, and sub-links.
    """
    try:
        response = requests.get(url, timeout=10.0)
        response.raise_for_status()
        html = response.text
    except requests.RequestException as e:
        return {"error": f"Failed to retrieve the website content: {e}"}

    soup = BeautifulSoup(html, "html.parser")

    meta_properties = [
        "og:description",
        "og:site_name",
        "og:title",
        "og:type",
        "og:url",
        "description",
        "keywords",
        "author",
    ]
    meta = {}
    for property_name in meta_properties:
        tag = soup.find("meta", property=property_name) or soup.find(
            "meta", attrs={"name": property_name}
        )
        if tag:
            meta[property_name] = tag.get("content", "")

    for ignore_tag in soup(["script", "style"]):
        ignore_tag.decompose()

    title = soup.title.string.strip() if soup.title else ""
    content = soup.body.get_text(separator="\n") if soup.body else ""

    links = []
    for a in soup.find_all("a", href=True):
        link_url = urljoin(url, a["href"])
        links.append({"title": a.text.strip(), "link": link_url})

    content = re.sub(r"[\n\r\t]+", "\n", content)
    content = re.sub(r" +", " ", content)
    content = re.sub(r"[\n ]{3,}", "\n\n", content)
    content = content.strip()

    if len(content) > max_content_length:
        content = content[:max_content_length].rsplit(" ", 1)[0] + "..."

    return {"meta": meta, "title": title, "content": content, "sub_links": links}